__all__ = ('id_sort_key', 'instance_or_id_to_instance', 'instance_or_id_to_snowflake',
    'iterable_of_instance_or_id_to_instances', 'iterable_of_instance_or_id_to_snowflakes', 'maybe_snowflake',
    'maybe_snowflake_pair', 'maybe_snowflake_token_pair')


def id_sort_key(entity):
    """
    Sort key for ``DiscordEntity`` instances.
    
    Parameters
    ----------
    entity : ``DiscordEntity``
        The discord entity to get identifier of.
    
    Returns
    -------
    entity_id : `int`
        The entity's identifier.
    """
    return entity.id


def instance_or_id_to_instance(obj, type_, name):
    """
    Converts the given `obj` to it's `type_` representation.
    
    Parameters
    ----------
    obj : `int`, `str` or`type_` instance
        The object to convert.
    type_ : `type` or (`tuple` of `type`)
        The type to convert.
    name : `str`
        The respective name of the object.
    
    Returns
    -------
    instance : `type_`
    
    Raises
    ------
    TypeError
        If `obj` was not given neither as `type_`, `str` or `int` instance.
    ValueError
        If `obj` was given as `str` or as `int` instance, but not as a valid snowflake, so `type_` cannot be precreated
        with it.
    
    Notes
    -----
    The given `type_` must have a `.precreate` function`.
    """
    obj_type = obj.__class__
    if issubclass(obj_type, type_):
        instance = obj
    else:
        if obj_type is int:
            snowflake = obj
        elif issubclass(obj_type, str):
            if 6 < len(obj) < 22 and obj.isdigit():
                snowflake = int(obj)
            else:
                raise ValueError(f'`{name}` was given as `str` instance, but not as a valid snowflake, got {obj!r}.')
        
        elif issubclass(obj_type, int):
            snowflake = int(obj)
        else:
            if type(type_) is tuple:
                type_name = ', '.join(t.__name__ for t in type_)
            else:
                type_name = type_.__name__
            
            raise TypeError(f'`{name}` can be given either as {type_name} instance, or as `int` or `str` representing '
                f'a snowflake, got {obj_type.__name__}.')
        
        if snowflake < 0 or snowflake > ((1<<64)-1):
            raise ValueError(f'`{name}` was given either as `int` or as `str` instance, but not as representing a '
                f'`uint64`, got {obj!r}.')
        
        if type(type_) is tuple:
            type_ = type_[0]
        
        instance = type_.precreate(snowflake)
    
    return instance


def iterable_of_instance_or_id_to_instances(iterable_obj, type_, name):
    """
    Converts the given `iterable_obj` to it's `type_`'s representation.
    
    Parameters
    ----------
    iterable_obj : `iterable` of `int`, `str` or`type_` instance
        The object to convert.
    type_ : `type` or (`tuple` of `type`)
        The type to convert.
    name : `str`
        The respective name of the object.
    
    Returns
    -------
    instances : `set` of `type_`
    
    Raises
    ------
    TypeError
        If `obj` was not given neither as `type_`, `str` or `int` instance.
    ValueError
        If `obj` was given as `str` or as `int` instance, but not as a valid snowflake, so `type_` cannot be precreated
        with it.
    
    Notes
    -----
    The given `type_` must have a `.precreate` function`.
    """
    iterator = getattr(type(iterable_obj), '__iter__', None)
    if iterator is None:
        raise TypeError(f'`{name}` can be `iterable`, got {iterable_obj.__class__.__name__}.')
    
    instances = set()
    
    for obj in iterator(iterable_obj):
        obj_type = obj.__class__
        if issubclass(obj_type, type_):
            instance = obj
        else:
            if obj_type is int:
                snowflake = obj
            elif issubclass(obj_type, str):
                if 6 < len(obj) < 22 and obj.isdigit():
                    snowflake = int(obj)
                else:
                    raise ValueError(f'`{name}` contains a `str` instance, but not as a valid snowflake, got {obj!r}.')
            
            elif issubclass(obj_type, int):
                snowflake = int(obj)
            else:
                if type(type_) is tuple:
                    type_name = ', '.join(t.__name__ for t in type_)
                else:
                    type_name = type_.__name__
                
                raise TypeError(f'`{name}` can contain either {type_name} instance, or an `int` or `str` representing '
                    f'a snowflake, got {obj_type.__name__}.')
            
            if snowflake < 0 or snowflake > ((1<<64)-1):
                raise ValueError(f'`{name}` contains an `int` or a `str` instance, but not as representing a '
                    f'`uint64`, got {obj!r}.')
            
            if type(type_) is tuple:
                type_ = type_[0]
        
            instance = type_.precreate(snowflake)
        
        instances.add(instance)
    
    return instances


def iterable_of_instance_or_id_to_snowflakes(iterable_obj, type_, name):
    """
    Converts the given `iterable_obj` to it's `type_`'s representation.
    
    Parameters
    ----------
    iterable_obj : `iterable` of `int`, `str` or`type_` instance
        The object to convert.
    type_ : `type` or (`tuple` of `type`)
        The type to convert.
    name : `str`
        The respective name of the object.
    
    Returns
    -------
    instances : `set` of `int`
    
    Raises
    ------
    TypeError
        If `obj` was not given neither as `type_`, `str` or `int` instance.
    ValueError
        If `obj` was given as `str` or as `int` instance, but not as a valid snowflake, so `type_` cannot be precreated
        with it.
    """
    iterator = getattr(type(iterable_obj), '__iter__', None)
    if iterator is None:
        raise TypeError(f'`{name}` can be `iterable`, got {iterable_obj.__class__.__name__}.')
    
    snowflakes = set()
    
    for obj in iterator(iterable_obj):
        obj_type = obj.__class__
        if issubclass(obj_type, type_):
            snowflake = obj.id
        else:
            if obj_type is int:
                snowflake = obj
            elif issubclass(obj_type, str):
                if 6 < len(obj) < 22 and obj.isdigit():
                    snowflake = int(obj)
                else:
                    raise ValueError(f'`{name}` contains a `str` instance, but not as a valid snowflake, got {obj!r}.')
            
            elif issubclass(obj_type, int):
                snowflake = int(obj)
            else:
                if type(type_) is tuple:
                    type_name = ', '.join(t.__name__ for t in type_)
                else:
                    type_name = type_.__name__
                
                raise TypeError(f'`{name}` can contain either {type_name} instance, or an `int` or `str` representing '
                    f'a snowflake, got {obj_type.__name__}.')
            
            if snowflake < 0 or snowflake > ((1<<64)-1):
                raise ValueError(f'`{name}` contains an `int` or a `str` instance, but not as representing a '
                    f'`uint64`, got {obj!r}.')
        
        snowflakes.add(snowflake)
    
    return snowflakes


def instance_or_id_to_snowflake(obj, type_, name):
    """
    Validates the given `obj` whether it is instance of the given `type_`, or is a valid snowflake representation.
    
    Parameters
    ----------
    obj : `int`, `str` or`type_` instance
        The object to validate.
    type_ : `type` of (`tuple` of `type`)
        Expected type.
    name : `str`
        The respective name of the object.
    
    Returns
    -------
    snowflake : `int`
    
    Raises
    ------
    TypeError
        If `obj` was not given neither as `type_`, `str` or `int` instance.
    ValueError
        If `obj` was given as `str` or as `int` instance, but not as a valid snowflake.
    
    Notes
    -----
    The given `type_`'s instances must have a `.id` attribute.
    """
    obj_type = obj.__class__
    if issubclass(obj_type, type_):
        snowflake = obj.id
    else:
        if obj_type is int:
            snowflake = obj
        elif issubclass(obj_type, str):
            if 6 < len(obj) < 22 and obj.isdigit():
                snowflake = int(obj)
            else:
                raise ValueError(f'`{name}` was given as `str` instance, but not as a valid snowflake, got {obj!r}.')
        
        elif issubclass(obj_type, int):
            snowflake = int(obj)
        else:
            if type(type_) is tuple:
                type_name = ', '.join(t.__name__ for t in type_)
            else:
                type_name = type_.__name__
            
            raise TypeError(f'`{name}` can be given either as {type_name} instance, or as `int` or `str` representing '
                f'a snowflake, got {obj_type.__name__}.')
        
        if (snowflake < 0) or (snowflake > ((1<<64)-1)):
            raise ValueError(f'`{name}` was given either as `int` or as `str` instance, but not as representing a '
                f'`uint64`, got {obj!r}.')
    
    return snowflake


def maybe_snowflake(value):
    """
    Converts the given `value` to `snowflake` if applicable. If not returns `None`.
    
    Parameters
    ----------
    value : `str`, `int` or `Any`
        A value what might be snowflake.
    
    Returns
    -------
    value : `None` or `int`
    
    Raises
    ------
    AssertionError
        - If `value` was passed as `str` and cannot be converted to `int`.
        - If the `value` is negative or it's bit length is over 64.
    """
    if isinstance(value, int):
        pass
    elif isinstance(value, str):
        if value.isdigit():
            if __debug__:
                value_length = len(value)
                if (value_length < 7) or (value_length > 21):
                    raise AssertionError('An `id` was given as `str` instance, but it\'s value is out of 64uint '
                        f'range, got {value!r}.')
            
            value = int(value)
        else:
            return None
    else:
        return None
    
    if __debug__:
        if (value < 0) or (value > ((1<<64)-1)):
            raise AssertionError('An `id` was given as `str` instance, but it\'s value is out of 64uint range, got '
                f'{value!r}.')
    
    return value


def maybe_snowflake_pair(value):
    """
    Checks whether the given value is a `tuple` of 2 snowflakes. If it, returns it, if not returns `None`.
    
    Parameters
    ----------
    value : `tuple` of (`str`, `int`) or `Any`
        A value what might be snowflake.
    
    Returns
    -------
    value : `tuple` (`int`, `int`) or `None`
    
    Raises
    ------
    AssertionError
        - If `value` contains a `str` element, what cannot be converted to `int`.
        - If `value` contains a value, what is negative or it's bit length is over 64.
    """
    if isinstance(value, tuple):
        if len(value) == 2:
            value_1, value_2 = value
            value_1 = maybe_snowflake(value_1)
            if value_1 is None:
                value = None
            else:
                value_2 = maybe_snowflake(value_2)
                if value_2 is None:
                    value = None
                else:
                    value = (value_1, value_2)
        else:
            value = None
    else:
        value = None
    
    return value


def maybe_snowflake_token_pair(value):
    """
    Checks whether the given value is a `tuple` of `2` elements: an identifier and a token. If not, returns `None`.
    
    Parameters
    ----------
    value : `tuple` of (`str`, `int`) or `Any`
        A value what might be snowflake.
    
    Returns
    -------
    value : `tuple` (`str`, `int`) or `None`
    
    Raises
    ------
    AssertionError
        - If `value` contains a `str` element, what cannot be converted to `int`.
        - If `value` contains a value, what is negative or it's bit length is over 64.
    """
    if isinstance(value, tuple):
        if len(value) == 2:
            value_1, value_2 = value
            value_1 = maybe_snowflake(value_1)
            if value_1 is None:
                value = None
            else:
                if isinstance(value_2, str):
                    value = (value_1, value_2)
                else:
                    value = None
        else:
            value = None
    else:
        value = None
    
    return value



__all__ = ('ApplicationExecutable', 'ApplicationSubEntity', 'EULA', 'ThirdPartySKU', )

from ..bases import DiscordEntity
from ..core import EULAS

class ApplicationSubEntity(DiscordEntity):
    """
    An un-typed entity stored inside of an ``Application``, as one of it's `.developers`, or `.publishers`.
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the entity.
    name : `str`
        The name of the entity.
    """
    __slots__ = ('name', )
    
    def __init__(self, data):
        """
        Creates a new ``ApplicationSubEntity`` instance.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Developers or Publisher data.
        """
        self.id = int(data['id'])
        self.name = data['name']
    
    def __repr__(self):
        """Returns the entity's representation."""
        return f'<{self.__class__.__name__} {self.name!r}, id={self.id}>'
    
    def __eq__(self, other):
        """Returns whether the two entities equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.id != other.id:
            return False
        
        if self.name != other.name:
            return False
        
        return True

class ApplicationExecutable:
    """
    Represents a game's executable.
    
    Attributes
    ----------
    parameters : `None` or `str`
        The parameters to start the application with. Defaults to `None`.
    is_launcher : `bool`
        Whether the application is a launcher. Defaults to `False`.
    name : `str`
        The executable's name.
    os : `str`
        The operation system, the executable is for.
    """
    __slots__ = ('parameters', 'is_launcher', 'name', 'os')
    
    def __init__(self, data):
        """
        Creates a new application executable with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Executable data.
        """
        self.name = data['name']
        self.os = data['os']
        self.parameters = data.get('parameters', None)
        self.is_launcher = data.get('is_launcher', False)
    
    def __repr__(self):
        """Returns the executable's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
            ' ',
            repr(self.name),
            ', os=',
            repr(self.os),
        ]
        
        parameters = self.parameters
        if (parameters is not None):
            repr_parts.append(', parameters=')
            repr_parts.append(repr(parameters))
        
        is_launcher = self.is_launcher
        if is_launcher:
            repr_parts.append(', is_launcher=True')
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    def __eq__(self, other):
        """Returns whether the two executables are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.name != other.name:
            return False
        
        if self.os != other.os:
            return False
        
        if self.parameters != other.parameters:
            return False
        
        if self.is_launcher != other.is_launcher:
            return False
        
        return True
    
    def __hash__(self):
        """Returns the entity's hash."""
        result = hash(self.name) ^ hash(self.os)
        
        parameters = self.parameters
        if (parameters is not None):
            result ^= hash(parameters)
        
        if self.is_launcher:
            result ^= (1<<15)
        
        return result

class ThirdPartySKU:
    """
    Represents a third party Stock Keeping Unit.
    
    distributor : `str`
        The distributor of the SKU.
    id : `str`
        The identifier of the third party SKU.
    sku : `str`
        Might be same as `.id`.
    """
    __slots__ = ('distributor', 'id', 'sku',)
    
    def __init__(self, data):
        """
        creates a new third party SKU object from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            SKU data.
        """
        self.distributor = data['distributor']
        self.id = data['id']
        self.sku = data['sku']
    
    def __repr__(self):
        """Returns the SKU's representation."""
        return f'<{self.__class__.__name__} distributor={self.distributor!r}, id={self.id!r}, sku={self.sku!r}>'
    
    def __eq__(self, other):
        """Returns whether the two SKU-s are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.distributor != other.distributor:
            return False
        
        if self.id != other.id:
            return False
        
        if self.sku != other.sku:
            return False
        
        return True
    
    def __hash__(self):
        """Returns the sku's hash."""
        result = hash(self.distributor)
        
        id_ = self.id
        result ^= hash(id_)
        
        sku = self.sku
        if sku != id_:
            result ^= hash(sku)
        
        return result

class EULA(DiscordEntity, immortal=True):
    """
    Represents a Discord end-user license agreement
    
    Attributes
    ----------
    id : `int`
        The unique identifier number of the eula.
    content : `str`
        The eula's content.
    name : `str`
        The eula's name.
    
    Notes
    -----
    The instances of the class support weakreferencing.
    """
    __slots__ = ('id', 'content', 'name')
    def __new__(cls, data):
        """
        Creates a new eula instance from the given parameters.
        
        If the eula already exists, returns that instead.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Eula data.
        """
        eula_id = int(data['id'])
        
        try:
            self = EULAS[eula_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = eula_id
            self._update_attributes(data)
            
            EULAS[eula_id] = self
        
        return self
    
    def _update_attributes(self, data):
        """
        Updates the eula with the received data from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        """
        self.content = data['content']
        self.name = data['name']
    
    def __repr__(self):
        """Returns the eula's representation"""
        return f'<{self.__class__.__name__} {self.name!r}, id={self.id}>'



__all__ = ()

import json, re, codecs
from http.cookies import SimpleCookie, CookieError, Morsel
from hashlib import md5, sha1, sha256

try:
    import ssl as module_ssl
except ImportError:
    module_ssl = None

try:
    import cchardet as chardet
except ImportError:
    try:
        import chardet
    except ImportError as err:
        message = ('chardet (or cchardet) is not installed, please make sure it is installed before importing '
            f'{__spec__.parent}')
        
        err.args = (message,)
        err.msg = message
        raise err from None

from .utils import imultidict
from .futures import Task, CancelledError
from .headers import METHOD_POST_ALL, METHOD_CONNECT, SET_COOKIE, CONTENT_LENGTH, CONNECTION, ACCEPT, ACCEPT_ENCODING, \
    HOST, TRANSFER_ENCODING, COOKIE, CONTENT_ENCODING, AUTHORIZATION, CONTENT_TYPE
from .helpers import BasicAuth
from .multipart import MimeType, create_payload
from .formdata import Formdata
from .protocol import HTTPStreamWriter

json_re = re.compile(r'^application/(?:[\w.+-]+?\+)?json')

class Fingerprint:
    """
    HTTP fingerprinting can be used to automate information systems and security audits. Automated security testing
    tools can use HTTP fingerprinting to narrow down the set of tests required, based on the specific platform or the
    specific web server being audited.
    
    Attributes
    ----------
    fingerprint : `bytes`
        The fingerprint's value.
    hash_function : `function`
        Hash function used by the fingerprint.
    
    Class Attributes
    ----------------
    HASH_FUNCTION_BY_DIGEST_LENGTH : `dict` of (`int`, `function`) items
        `fingerprint`'s length - `hash-function` relation mapping.
    """
    __slots__ = ('fingerprint', 'hash_function',)
    
    HASH_FUNCTION_BY_DIGEST_LENGTH = {
        16: md5,
        20: sha1,
        32: sha256,
    }
    
    def __new__(cls, fingerprint):
        """
        Creates a new ``Fingerprint`` instance with the given parameters.
        
        Parameters
        ----------
        fingerprint : `bytes`
            Fingerprint value.
        
        Raises
        ------
        ValueError
            - If `fingerprint`'s length is not any of the expected ones.
            - If the detected `hash_function` is `md5` or `sha1`.
        """
        fingerprint_length = len(fingerprint)
        
        try:
            hash_function = cls.HASH_FUNCTION_BY_DIGEST_LENGTH[fingerprint_length]
        except KeyError:
            raise ValueError(f'`fingerprint` has invalid length, got {fingerprint_length!r}, {fingerprint!r}') from None
        
        if hash_function is md5 or hash_function is sha1:
            raise ValueError('`md5` and `sha1` are insecure and not supported, use `sha256`.')
        
        self = object.__new__(cls)
        self.hash_function = hash_function
        self.fingerprint = fingerprint
        return self

    def check(self, transport):
        """
        Checks whether the given transport's ssl data matches the fingerprint.
        
        Parameters
        ----------
        transport : `Any`
            Asynchronous transport implementation.
        
        Raises
        ------
        ValueError
            If the fingerprint don't match.
        """
        if transport.get_extra_info('sslcontext') is None:
            return
        
        ssl_object = transport.get_extra_info('ssl_object')
        cert = ssl_object.getpeercert(binary_form=True)
        received = self.hash_function(cert).digest()
        fingerprint = self.fingerprint
        if received == fingerprint:
            return
            
        host, port, *_ = transport.get_extra_info('peername')
        raise ValueError(f'The expected fingerprint: {fingerprint!r} not matches the received; received={received!r}, '
            f'host={host!r}, port={port!r}.')

if module_ssl is None:
    SSL_ALLOWED_TYPES = (type(None), )
else:
    SSL_ALLOWED_TYPES = (module_ssl.SSLContext, bool, Fingerprint, type(None))


class ConnectionKey:
    """
    Contains information about a host, like proxy, TLS to prevent reusing wrong connections from the pool.
    
    Attributes
    ----------
    host : `str`
        The host's ip address.
    is_ssl : `bool`
        Whether the connection is secure.
    port : `int`
        The host's port.
    proxy_auth : `None` or ``BasicAuth``
        Proxy authorization.
    proxy_url : `None` or ``URL``
        Proxy's url.
    ssl : `None`, ``SSLContext``, `bool`, ``Fingerprint``
        The connection's ssl type.
    """
    __slots__ = ('host', 'is_ssl', 'port', 'proxy_auth',  'proxy_url', 'ssl',) # + 'proxy_header_hash',
    
    def __init__(self, request):
        # proxy_headers = request.proxy_headers
        # if request.proxy_headers is not None:
        #     proxy_header_hash = hash(tuple(proxy_headers.items()))
        # else:
        #     proxy_header_hash = None
        
        self.host = request.host
        self.port = request.port
        self.is_ssl = request.is_ssl()
        self.ssl = request.ssl
        self.proxy_auth = request.proxy_auth
        self.proxy_url = request.proxy_url
        # self.proxy_header_hash = proxy_header_hash
    
    def __repr__(self):
        """Returns the connection key's representation."""
        return f'<{self.__class__.__name__} host={self.host!r}, port={self.port!r}>'
    
    def __eq__(self, other):
        """Returns whether the two connection keys are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.host != other.host:
            return False
        
        if self.port != other.port:
            return False
        
        if self.is_ssl != other.is_ssl:
            return False
        
        if self.ssl is None:
            if other.ssl is not None:
                return False
        else:
            if other.ssl is None:
                return False
            
            if self.ssl != other.ssl:
                return False
            
        if self.proxy_auth is None:
            if other.proxy_auth is not None:
                return False
        else:
            if other.proxy_auth is None:
                return False
            
            if self.proxy_auth != other.proxy_auth:
                return False

        if self.proxy_url is None:
            if other.proxy_url is not None:
                return False
        else:
            if other.proxy_url is None:
                return False
            
            if self.proxy_url != other.proxy_url:
                return False
        
        return True
    
    def __hash__(self):
        """Returns the connection key's hash value."""
        return hash(self.host) ^ (self.port << 17) ^ hash(self.is_ssl) ^ hash(self.ssl) ^ hash(self.proxy_auth) ^ \
               hash(self.proxy_url)

class RequestInfo:
    """
    Base information representing a request.
    
    Attributes
    ----------
    headers : ``imultidict``
        The respective request's headers.
    method : `str`
        The respective request's method.
    real_url : ``URL``
        The url given to request.
    url : ``URL``
        The requested url without fragments. Can be same as ``.real_url``.
    """
    __slots__ = ('headers', 'method', 'real_url', 'url',)
    def __init__(self, request):
        """
        Creates a new ``RequestInfo`` instance representing the given request.
        
        Parameters
        ----------
        request : ``ClientRequest``
            The represented request.
        """
        self.url = request.url
        self.method = request.method
        self.headers = request.headers
        self.real_url = request.original_url
    
    def __repr__(self):
        """Returns the request info's representation."""
        return f'<{self.__class__.__name__} url={self.url!r}>'


DEFAULT_HEADERS = (
    (ACCEPT, '*/*'),
    (ACCEPT_ENCODING, 'gzip, deflate'),
)

class ClientRequest:
    """
    Http request class used by ``HTTPClient``.
    
    Attributes
    ----------
    auth : `None` or ``BasicAuth``
        Authorization sent with the request.
    body : `None`, ``PayloadBase`` instance
        The request's body.
    chunked : `bool`
        Whether the request is sent chunked.
    compression : `None` or `str`
        Compression used when sending the request.
    headers : `imultidict`
        The headers of the request.
    loop : ``EventThread``
        The event loop, trough what the request is executed.
    method : `str`
        The request's method.
    original_url : ``URL``
        The original url, what was asked to request.
    proxy_auth : `None` or ``BasicAuth``
        Proxy authorization sent with the request.
    proxy_url : `None` or ``URL``
        Proxy url to use if applicable.
    response : `None` or ``ClientResponse``
        Object representing the received response. Set as `None` till ``.send`` finishes.
    ssl : `None` `None`, ``SSLContext``, `bool`, ``Fingerprint``
        The connection's ssl type.
    url : ``URL``
        The url, what will be requested.
    writer : `None` or ``Task`` of ``.write_bytes``
        Payload writer task, what is present meanwhile the request's payload is sending.
    """
    __slots__ = ('auth', 'body', 'chunked', 'compression', 'headers', 'loop', 'method', 'original_url', 'proxy_auth',
        'proxy_url', 'response', 'ssl', 'url', 'writer',)
    
    def __new__(cls, loop, method, url, headers, data, params, cookies, auth, proxy_url, proxy_auth, ssl):
        """
        Creates a new ``ClientRequest`` instance with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop, trough what the request is executed.
        method : `str`
            The request's method.
        url : ``URL``
            The url to request.
        headers : `None`, `dict` or ``imultidict``
            Headers of the request.
        data : `None`, `bytes-like`, `io-like`, ``Formdata`
            Data to send as the request's body.
        params : `dict` of (`str`, (`str`, `int`, `float`, `bool`)) items
            Query string parameters.
        cookies : `None` or ``CookieJar``
            Cookies OwO.
        auth : `None` or ``BasicAuth``
            Authorization sent with the request.
        proxy_url : `None` or ``URL``
            Proxy url to use if applicable.
        proxy_auth : `None` or ``BasicAuth``
            Proxy authorization sent with the request.
        ssl : `None` `None`, ``SSLContext``, `bool`, ``Fingerprint``
            The connection's ssl type.
        
        Raises
        ------
        TypeError
            - `proxy_auth`'s type is incorrect.
            - ˙Cannot serialize a field of the given `data`.
        ValueError
            - Host could not be detected from `url`.
            - The `proxy_url`'s scheme is not `http`.
            - `compression` and `Content-Encoding` would be set at the same time.
            - `chunked` cannot be set, because `Transfer-Encoding: chunked` is already set.
            - `chunked` cannot be set, because `Content-Length` header is already present.
        RuntimeError
            - If one of `data`'s field's content has unknown content-encoding.
            - If one of `data`'s field's content has unknown content-transfer-encoding.
        """
        # Convert headers
        headers = imultidict(headers)
        
        # Add extra query parameters to the url and remove fragments
        url = url.extend_query(params)
        request_url = url.with_fragment(None)
        
        if not url.host:
            raise ValueError('Host could not be detected.')
        
        # Check authorization
        if auth is None:
            # If authorization is given, try to detect from url.
            username = url.user
            password = url.password
            
            if (username is not None) and username:
                if password is None:
                    password = ''
                
                auth = BasicAuth(username, password)
        
        # Store auth in headers is applicable.
        if (auth is not None):
            headers[AUTHORIZATION] = auth.encode()
        
        for key, value in DEFAULT_HEADERS:
            headers.setdefault(key, value)
        
        # Add host to headers if not present.
        if HOST not in headers:
            netloc = request_url.raw_host
            if not request_url.is_default_port():
                netloc = f'{netloc}:{request_url.port}'
            
            headers[HOST] = netloc
        
        # Update cookies
        if (cookies is not None) and cookies:
            cookie = SimpleCookie()
            if COOKIE in headers:
                cookie.load(headers.get(COOKIE, ''))
                del headers[COOKIE]
            
            for key, value in cookies.items():
                if isinstance(key, Morsel):
                    # Preserve coded_value
                    try:
                        morsel_value = value.get(value.key, None)
                    except KeyError:
                        morsel_value = Morsel()
                    
                    morsel_value.set(value.key, value.value, value.coded_value)
                    value = morsel_value
                
                cookie[key] = value
            
            headers[COOKIE] = cookie.output(header='', sep=';').strip()
        
        # Check proxy settings.
        if proxy_url is not None:
            if proxy_url.scheme != 'http':
                raise ValueError(f'Only http proxies are supported, got {proxy_url!r}.')
            
            if (proxy_auth is not None):
                proxy_auth_type = proxy_auth.__class__
                if proxy_auth_type is not BasicAuth:
                    raise TypeError(f'`proxy_auth` must be `None` or `{BasicAuth.__name__}`, got '
                        f'{proxy_auth_type.__name__}.')
        
        # Needed for transfer data checks
        chunked = True
        compression = headers.get(CONTENT_ENCODING, None)
        
        # Get request content encoding.
        if (data is not None):
            if data:
                if (compression is not None):
                    if headers.get(CONTENT_ENCODING, ''):
                        raise ValueError('Compression can not be set if `Content-Encoding` header is set.')
                    
                    chunked = True
                
                # formdata
                if isinstance(data, Formdata):
                    data = data()
                else:
                    try:
                        data = create_payload(data, {'disposition': None})
                    except LookupError:
                        data = Formdata.from_fields(data)()
                
                if not chunked:
                    if CONTENT_LENGTH not in headers:
                        size = data.size
                        if size is None:
                            chunked = True
                        else:
                            if CONTENT_LENGTH not in headers:
                                headers[CONTENT_LENGTH] = str(size)
                
                if CONTENT_TYPE not in headers:
                    headers[CONTENT_TYPE] = data.content_type
                
                data_headers = data.headers
                if data_headers:
                    for key, value in data_headers.items():
                        headers.setdefault(key, value)
            else:
                data = None
        
        # Analyze transfer-encoding header.
        transfer_encoding = headers.get(TRANSFER_ENCODING, '').lower()
        
        if 'chunked' in transfer_encoding:
            if chunked:
                raise ValueError('Chunked can not be set if `Transfer-Encoding: chunked` header is already set.')
        
        elif chunked:
            if CONTENT_LENGTH in headers:
                raise ValueError('Chunked can not be set if `Content-Length` header is set.')
            headers[TRANSFER_ENCODING] = 'chunked'
        
        else:
            if CONTENT_LENGTH not in headers:
                headers[CONTENT_LENGTH] = '0' if data is None else str(len(data))
        
        # Set default content-type.
        if (method in METHOD_POST_ALL) and (CONTENT_TYPE not in headers):
            headers[CONTENT_TYPE] = 'application/octet-stream'
        
        # Everything seems correct, create the object.
        self = object.__new__(cls)
        self.original_url = url
        self.url = request_url
        self.method = method
        self.loop = loop
        self.ssl = ssl
        self.chunked = chunked
        self.compression = compression
        self.body = data
        self.auth = auth
        self.writer = None
        self.response = None
        self.headers = headers
        self.proxy_url = proxy_url
        self.proxy_auth = proxy_auth
        
        return self
    
    def is_ssl(self):
        """
        Returns whether the request is ssl.
        
        Returns
        -------
        is_ssl : `bool`
        """
        return self.url.scheme in ('https', 'wss')
    
    @property
    def connection_key(self):
        """
        Returns the connection key of request.
        
        Returns
        -------
        connection_key : ``ConnectionKey``
        """
        return ConnectionKey(self)
    
    @property
    def request_info(self):
        """
        Returns base information representing the request.
        
        Returns
        -------
        request_info : ``RequestInfo``
        """
        return RequestInfo(self)
    
    @property
    def host(self):
        """
        Returns the request's host.
        
        Returns
        -------
        host : `str`
        """
        return self.url.host
    
    @property
    def port(self):
        """
        Returns the request's port.
        
        Returns
        -------
        port : `int`
        """
        return self.url.port
    
    async def write_bytes(self, writer, connection):
        """
        Writes the request's body..
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Writer used to write the request's body into the connection's transport.
        connection : ``Connection``
            Connection of the request with what the payload is sent.
        """
        # Support coroutines that yields bytes objects.
        try:
            body = self.body
            if (body is not None):
                await self.body.write(writer)
            await writer.write_eof()
        except OSError as err:
            new_err = OSError(err.errno, f'Can not write request body for {self.url!r}.')
            new_err.__context__ = err
            new_err.__cause__ = err
            connection.protocol.set_exception(new_err)
        except CancelledError as err:
            if not connection.closed:
               connection.protocol.set_exception(err)
        except BaseException as err:
            connection.protocol.set_exception(err)
            raise
        finally:
            self.writer = None
    
    def send(self, connection):
        """
        Sends the request.
        
        Parameters
        ----------
        connection : ``Connection``
            Connection, what is used to send the request.
        
        Returns
        -------
        response : `coroutine` of ``ClientResponse.start`` ->
        """
        try:
            url = self.url
            if self.method == METHOD_CONNECT:
                path = f'{url.raw_host}:{url.port}'
            elif (self.proxy_url is not None) and (not self.is_ssl()):
                path = str(url)
            else:
                path = url.raw_path
                if url.raw_query_string:
                    path = f'{path}?{url.raw_query_string}'
            
            protocol = connection.protocol
            writer = HTTPStreamWriter(protocol, self.compression, self.chunked)
            
            protocol.write_http_request(self.method, path, self.headers)
            
            self.writer = Task(self.write_bytes(writer, connection), self.loop)
            
            self.response = response = ClientResponse(self, connection)
            
            return response.start()
        
        except:
            connection.close()
            raise
    
    def terminate(self):
        """
        Terminates the request's writing task if applicable.
        """
        writer = self.writer
        if (writer is not None):
            self.writer = None
            writer.cancel()


class ClientResponse:
    """
    Http response class used by ``HTTPClient``.
    
    Attributes
    ----------
    _released : `bool`
        Whether the connection is released.
    body : `None` or `bytes`
        The received response body. Set as `None` if the response body is not yet received, or if it is empty.
    closed : `bool`
        Whether the response is closed.
    connection : `None` or ``Connection``
        Connection used to receive the request response. Set as `None` if the response is ``.close``-d or
        ``.release``-d.
    payload_waiter : `None` or ``Future``
        Future used to retrieve the response's body. It's result is set, when the respective protocol's reader task
        finished.
    cookies : `http.cookies.SimpleCookie`
        Received cookies with the response.
    headers : `None` or ``imultidict``
        Headers of the response. Set when the http response is successfully received.
    history : `None` or `tuple` of ``ClientResponse``
        Response history. Set as `tuple` of responses from outside.
    loop : ``EventThread``
        The event loop, trough what the request is executed.
    method : `str`
        Method of the respective request.
    status : `None` or `int`
        Received status code. Set as `0` by default.
    url : ``URL``
        The requested url.
    writer : ``Task`` of ``ClientRequest.write_bytes``
        Payload writer task of the respective request.
    raw_message : `None` or ``RawResponseMessage``
        Raw received http response.
    """
    __slots__ = ('_released', 'body', 'closed', 'connection', 'payload_waiter', 'cookies', 'headers', 'history', 'loop',
        'method', 'status', 'url', 'writer', 'raw_message')
       
    def __new__(cls, request, connection):
        """
        Crates a new ``ClientResponse`` instance from the given request and connection.
        
        Parameters
        ----------
        request : ``ClientRequest``
            The respective request.
        connection : ``Connection``
            The connection used to send the request and receive the response.
        """
        self = object.__new__(cls)
        self.loop = request.loop
        self.method = request.method
        self.url = request.original_url
        
        self.writer = request.writer
        self.closed = False
        self.cookies = SimpleCookie()
        self._released = False
        
        self.body = None
        self.status = 0
        self.payload_waiter = None
        self.headers = None
        self.connection = connection
        
        self.raw_message = None
        self.history = None  # will be added later
        
        return self
    
    @property
    def reason(self):
        """
        Returns the server response reason.
        
        Returns
        -------
        reason : `None` or `str`
        """
        message = self.raw_message
        if (message is not None):
            reason = message.reason
            if (reason is not None):
                return reason.decode()
    
    def __del__(self):
        """releases the response if not yet closed."""
        if self.closed:
            return
        
        self._release_connection()
    
    def __repr__(self):
        """Returns the response's representation."""
        ascii_encodable_url = str(self.url)
        
        return f'<{self.__class__.__name__}({ascii_encodable_url}) [{self.status} {self.reason!r}]>'
    
    async def start(self,):
        """
        Starts response processing.
        
        This method is a coroutine.
        
        Returns
        -------
        self : ``ClientResponse``
        """
        try:
            protocol = self.connection.protocol
            
            payload_waiter = protocol.set_payload_reader(protocol._read_http_response())
            self.raw_message = message = await payload_waiter
            protocol.handle_payload_waiter_cancellation()
            payload_reader = protocol.get_payload_reader_task(message)
            if (payload_reader is None):
                payload_waiter = None
                self._response_eof(None)
            else:
                payload_waiter = protocol.set_payload_reader(payload_reader)
                payload_waiter.add_done_callback(self._response_eof)
            
            # response status
            self.status = message.status
            # headers
            self.headers = message.headers
            # OwO
            self.payload_waiter = payload_waiter
            
            # cookies
            for header in self.headers.get_all(SET_COOKIE, ()):
                try:
                    self.cookies.load(header)
                except CookieError: # so sad
                    pass
        except:
            self.close()
            raise
        
        return self
    
    def _response_eof(self, future):
        """
        Future callback added to the payload waiter future, to release the used connection.
        
        Parameters
        ----------
        future : ``Future``
            ``.payload_waiter`` future.
        """
        if self.closed:
            return
        
        self.payload_waiter = None
        
        connection = self.connection
        if (connection is not None):
            # Websocket, protocol could be `None`, because connection could be detached.
            if (connection.protocol is not None) and self.raw_message.upgraded:
                return
            
            self._release_connection()
        
        self.closed = True
        self._cleanup_writer()
    
    def _release_connection(self):
        """
        Releases the response's connection.
        
        If the connection type is "close", closes the protocol as well.
        """
        connection = self.connection
        if connection is None:
            return
        
        headers = self.headers
        if (headers is not None):
            try:
                connection_type = headers[CONNECTION]
            except KeyError:
                pass
            else:
                if connection_type == 'close':
                    protocol = connection.protocol
                    if (protocol is not None):
                        protocol.close()
        
        connection.release()
        self.connection = None
    
    def _notify_content(self):
        """
        Called when response reading is cancelled or released. Sets `ConnectionError` to the respective protocol if
        the payload is still reading.
        """
        payload_waiter = self.payload_waiter
        if (payload_waiter is not None):
            connection = self.connection
            if (connection is not None):
                connection.protocol.set_exception(ConnectionError('Connection closed.'))
        
        self._released = True
    
    def _cleanup_writer(self):
        """
        Cancels the writer task of the respective request. Called when the response is cancelled or released, or if
        reading the whole response is done.
        """
        writer = self.writer
        if (writer is not None):
            self.writer = None
            writer.cancel()
    
    async def read(self):
        """
        Reads the response's body.
        
        This method is a coroutine.
        
        Returns
        -------
        body : `bytes`
        """
        payload_waiter = self.payload_waiter
        if (payload_waiter is None):
            body = self.body
        else:
            try:
                self.body = body = await payload_waiter
            finally:
                self.payload_waiter = None
        
        return body
    
    def get_encoding(self):
        """
        Gets the encoding of the response's body.
        
        Returns
        -------
        encoding : `str`
            Defaults to `'utf-8'`.
        """
        content_type = self.headers.get(CONTENT_TYPE, '').lower()
        mime_type = MimeType(content_type)
        
        encoding = mime_type.parameters.get('charset', None)
        if encoding is not None:
            try:
                codecs.lookup(encoding)
            except LookupError:
                encoding = None
        
        if encoding is None:
            if mime_type.type == 'application' and mime_type.sub_type == 'json':
                encoding = 'utf-8' # RFC 7159 states that the default encoding is UTF-8.
            else:
                encoding = chardet.detect(self.body)['encoding']
        
        if not encoding:
            encoding = 'utf-8'
        
        return encoding
    
    async def text(self, encoding=None, errors='strict'):
        """
        Loads the response's content as text.
        
        This method is a coroutine.
        
        Parameters
        ----------
        encoding : `None` or `str`, Optional
            If no encoding is given, then detects it from the payload-
        errors : `str`, Optional
            May be given to set a different error handling scheme. The default `errors` value is `'strict'`, meaning
            that encoding errors raise a `UnicodeError`. Other possible values are `'ignore'`, `'replace'`,
            `'xmlcharrefreplace'`, `'backslashreplace'` and any other name registered via `codecs.register_error()`.
        
        Returns
        -------
        text : `str`
        """
        body = await self.read()
        if body is None:
            return
        
        if encoding is None:
            encoding = self.get_encoding()
        
        return body.decode(encoding, errors)
    
    async def json(self, encoding=None, loader=json.loads, content_type='application/json'):
        """
        Loads the response's content as a json.
        
        This method is a coroutine.
        
        Parameters
        ----------
        encoding : None` or `str`, Optional
            Encoding to use instead of the response's. If given as `None` (so by default), then will use the response's
            own encoding.
        loader : `callable`, Optional
            Json loader. Defaults to json.loads`.
        content_type : `str`, Optional
            Content type to use instead of the default one. Defaults to `'application/json'`.
        
        Returns
        -------
        json : `Any`
        
        Raises
        ------
        TypeError
            If the response's mime_type do not match.
        """
        body = await self.read()
        if body is None:
            return
        
        if content_type is not None:
            received_content_type = self.headers.get(CONTENT_TYPE, '').lower()
            
            if (json_re.match(received_content_type) is None) if (content_type == 'application/json') else \
                    (content_type not in received_content_type):
                raise TypeError(f'Attempt to decode JSON with unexpected mime_type: {received_content_type!r}.')
        
        stripped = body.strip()
        if not stripped:
            return None
        
        if encoding is None:
            encoding = self.get_encoding()
        
        return loader(stripped.decode(encoding))
    
    async def __aenter__(self):
        """
        Enters the client response as an asynchronous context manager.
        
        This method is a coroutine.
        """
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Releases the response if not yet closed.
        
        This method is a coroutine.
        """
        self.release()
        return False

    def close(self):
        """
        Closes the response and it's connection. The used connection will not be reused after.
        """
        if not self._released:
            self._notify_content()
        
        if self.closed:
            return
        
        self.closed = True
        
        connection = self.connection
        if (connection is not None):
            self.connection = None
            connection.close()
        
        self._cleanup_writer()
    
    def release(self):
        """
        Releases the response and it's connection. The used connection might be reused after.
        """
        if not self._released:
            self._notify_content()
        
        if self.closed:
            return
        
        self.closed = True
        
        self._release_connection()
        self._cleanup_writer()



__all__ = ('GuildUserChunkEvent', 'VoiceServerUpdateEvent',)

from ...env import CACHE_PRESENCE

from ...backend.utils import set_docs

from ..bases import EventBase
from ..guild import Guild
from ..user import User
from ..core import GUILDS


class GuildUserChunkEvent(EventBase):
    """
    Represents a processed `GUILD_MEMBERS_CHUNK` dispatch event.
    
    Attributes
    ----------
    guild_id : `int`
        The guild's identifier, what received the user chunk.
    users : `list` of ``ClientUserBase``
        The received users.
    nonce : `None` or `str`
        A nonce to identify guild user chunk response.
    index : `int`
        The index of the received chunk response (0 <= index < count).
    count : `int`
        The total number of chunk responses what Discord sends for the respective gateway.
    """
    __slots__ = ('guild_id', 'users', 'nonce', 'index', 'count')
    
    if CACHE_PRESENCE:
        def __new__(cls, data):
            guild_id = int(data['guild_id'])
            guild = GUILDS.get(guild_id, None)
            
            users = []
            for user_data in data['members']:
                user = User(user_data, guild)
                users.append(user)
            
            try:
                presence_datas = data['presences']
            except KeyError:
                pass
            else:
                if (guild is not None):
                    guild._apply_presences(presence_datas)
            
            self = object.__new__(GuildUserChunkEvent)
            self.guild_id = guild_id
            self.users = users
            self.nonce = data.get('nonce', None)
            self.index = data.get('chunk_index', 0)
            self.count = data.get('chunk_count', 1)
            
            return self
    
    else:
        def __new__(cls, data):
            guild_id = int(data['guild_id'])
            guild = GUILDS.get(guild_id, None)
            
            users = []
            for user_data in data['members']:
                user = User(user_data, guild)
                users.append(user)
            
            self = object.__new__(GuildUserChunkEvent)
            self.guild = guild
            self.users = users
            self.nonce = data.get('nonce', None)
            self.index = data.get('chunk_index', 0)
            self.count = data.get('chunk_count', 1)
            
            return self
    
    set_docs(__new__,
        """
        Creates a new guild user chunk event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Guild user chunk event data.
        """
    )
    
    def __repr__(self):
        """Returns the representation of the guild user chunk event."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' guild_id=')
        repr_parts.append(repr(self.guild_id))
        
        count = self.count
        if count != 1:
            repr_parts.append(', index=')
            repr_parts.append(repr(self.index))
            repr_parts.append('/')
            repr_parts.append(repr(count))
        
        repr_parts.append(', user count=')
        repr_parts.append(repr(len(self.users)))
        
        nonce = self.nonce
        if (nonce is not None):
            repr_parts.append(' nonce=')
            repr_parts.append(repr(nonce))
        
        repr_parts.append('>')
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 5
    
    def __iter__(self):
        """
        Unpacks the guild user chunk event.
        
        This method is a generator.
        """
        yield self.guild_id
        yield self.users
        yield self.nonce
        yield self.index
        yield self.count


class VoiceServerUpdateEvent(EventBase):
    """
    Represents a `VOICE_SERVER_UPDATE` event.
    
    Attributes
    ----------
    endpoint : `None` or `str`
        The voice server's host.
    guild_id : `int`
        The respective guild's identifier.
    token : `str`
        Voice connection token.
    """
    __slots__ = ('endpoint', 'guild_id', 'token')
    
    def __new__(cls, data):
        """
        Creates a new voice server update event from the given voice server update data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Voice server update data.
        """
        guild_id = data.get('guild_id', None)
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.endpoint = data.get('endpoint', None)
        self.token = data.get('token', None)
        
        return self
    
    def __repr__(self):
        """Returns the representation of the voice server update event."""
        repr_parts = ['<', self.__class__.__name__,]
        
        repr_parts.append(' guild_id=')
        repr_parts.append(repr(self.guild_id))
        
        endpoint = self.endpoint
        if (endpoint is not None):
            repr_parts.append(', endpoint=')
            repr_parts.append(repr(endpoint))
        
        token = self.token
        if (token is not None):
            repr_parts.append(', token=')
            repr_parts.append(repr(self.token))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 3
    
    def __iter__(self):
        """
        Unpacks the voice server update event.
        
        This method is a generator.
        """
        yield self.guild_id
        yield self.endpoint
        yield self.token



__all__ = ('PrivacyLevel', 'ScheduledEventEntityType', 'ScheduledEventStatus',)

from ..bases import PreinstancedBase, Preinstance as P

from .metadata import StageEntityMetadata, LocationEntityMetadata

class ScheduledEventStatus(PreinstancedBase):
    """
    Represents a scheduled event's status.
    
    Attributes
    ----------
    name : `str`
        The name of the scheduled event status.
    value : `int`
        The identifier value the scheduled event status.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ScheduledEventStatus``) items
        Stores the predefined ``ScheduledEventStatus`` instances. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The scheduled event status' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the scheduled event statuses.
    
    Every predefined scheduled event can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | Name          | Value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | scheduled             | scheduled     | 1     |
    +-----------------------+---------------+-------+
    | active                | active        | 2     |
    +-----------------------+---------------+-------+
    | completed             | completed     | 3     |
    +-----------------------+---------------+-------+
    | cancelled             | cancelled     | 4     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    none = P(0, 'none')
    scheduled = P(1, 'scheduled')
    active = P(2, 'active')
    completed = P(3, 'completed')
    cancelled = P(4, 'cancelled')


class ScheduledEventEntityType(PreinstancedBase):
    """
    Represents a scheduled event's entity's type.
    
    Attributes
    ----------
    name : `str`
        The name of the scheduled event's entity's type.
    value : `int`
        The identifier value the scheduled event's entity type.
    metadata_type : `None` or ``ScheduledEventEntityMetadata`` subclass
        The scheduled event's metadata's applicable type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ScheduledEventEntityType``) items
        Stores the predefined ``ScheduledEventEntityType`` instances. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The scheduled event's entity's type's values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the scheduled event entity types.
    
    Every predefined scheduled event can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+-------------------------------+
    | Class attribute name  | Name          | Value | Metadata type                 |
    +=======================+===============+=======+===============================+
    | none                  | none          | 0     | None                          |
    +-----------------------+---------------+-------+-------------------------------+
    | stage                 | stage         | 1     | ``StageEntityMetadata``       |
    +-----------------------+---------------+-------+-------------------------------+
    | voice                 | voice         | 2     | `None`                        |
    +-----------------------+---------------+-------+-------------------------------+
    | location              | location      | 3     | ``LocationEntityMetadata``    |
    +-----------------------+---------------+-------+-------------------------------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ('metadata_type',)
    

    @classmethod
    def _from_value(cls, value):
        """
        Creates a scheduled event entity type from the given id and stores it at class's `.INSTANCES`.
        
        Called by `.get` when no scheduled event entity type was found with the given id.
        
        Parameters
        ----------
        id_ : `str`
            The identifier of the scheduled event entity type.
        
        Returns
        -------
        self : ``ScheduledEventEntityType``
        """
        self = object.__new__(cls)
        self.name = cls.DEFAULT_NAME
        self.metadata_type = None
        
        self.INSTANCES[value] = self
        return self
    
    
    def __init__(self, value, name, metadata_type):
        """
        Creates a new scheduled event entity type instance from the given parameters.
        
        Parameters
        ----------
        value : `str`
            The unique identifier of the scheduled event entity type.
        name : `str`
            The name of the scheduled event entity type.
        metadata_type : `None` or ``ScheduledEventEntityMetadata`` subclass
            The scheduled event's metadata's applicable type.
        """
        self.name = name
        self.value = value
        self.metadata_type = metadata_type
        self.INSTANCES[value] = self
    
    
    none = P(0, 'none', None)
    stage = P(1, 'stage', StageEntityMetadata)
    voice = P(2, 'voice', None)
    location = P(3, 'location', LocationEntityMetadata)



class PrivacyLevel(PreinstancedBase):
    """
    Represents a stage channel's privacy level.
    
    Attributes
    ----------
    name : `str`
        The name of the privacy level.
    value : `int`
        The identifier value the privacy level.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``PrivacyLevel``) items
        Stores the predefined ``PrivacyLevel`` instances. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The privacy level' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the privacy levels.
    
    Every predefined privacy level can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | Name          | Value |
    +=======================+===============+=======+
    | none                  | none          | 0     |
    +-----------------------+---------------+-------+
    | public                | public        | 1     |
    +-----------------------+---------------+-------+
    | guild_only            | guild_only    | 2     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    none = P(0, 'none')
    public = P(1, 'public')
    guild_only = P(2, 'guild_only')



__all__ = ()

from ...discord.events.handling_helpers import asynclist
from ...discord.events.core import DEFAULT_EVENT_HANDLER

from ...ext.command_utils import MessageCreateWaitfor

from . import EXTENSION_SETUP_HOOKS
from .command_processor import CommandProcessor


def extension_setup_hook(client, command_processor):
    """
    Hook to merge `command_utils`'s `message_create` event with the command processor.
    
    Attributes
    ----------
    client : ``Client``
        The client the extension is setup.
    command_processor : ``CommandProcessor``
        The create command processor instance.
    
    Raises
    ------
    RuntimeError
        If the client already has a ``CommandProcessor` added to it.
    """
    event_message_create = client.events.message_create
    if (event_message_create is DEFAULT_EVENT_HANDLER):
        event = None
    elif type(event_message_create) is asynclist:
        for event in list.__iter__(event_message_create):
            if isinstance(event, CommandProcessor):
                raise RuntimeError(f'The client already has a `{CommandProcessor.__name__}` instance added as '
                    f'event.')
            
            if isinstance(event, MessageCreateWaitfor):
                break
        else:
            event = None
    else:
        event = event_message_create
        if isinstance(event, CommandProcessor):
            raise RuntimeError(f'The client already has a `{CommandProcessor.__name__}` instance added as '
                f'event.')
        
        if not isinstance(event, MessageCreateWaitfor):
            event = None
    
    if (event is not None):
        client.events.remove(event)
        
        command_processor.waitfors.update(event.waitfors)
        event.waitfors.clear()

EXTENSION_SETUP_HOOKS.append(extension_setup_hook)



__all__ = ('Embed', )

from ...backend.utils import copy_docs

from ..utils import timestamp_to_datetime, datetime_to_timestamp

from .embed_base import EmbedBase, EmbedFooter, EmbedImage, EmbedThumbnail, EmbedVideo, EmbedProvider, EmbedAuthor, \
    EmbedField

RICH_EMBED_FIELDS = frozenset((
    'author',
    'fields',
    'footer',
    'image',
    'provider',
    'thumbnail',
    'video',
))

class Embed(EmbedBase):
    """
    Represents Discord embedded content. There are two defined embed classes, the other one is ``EmbedCore``.
    
    Embeds are easier to build with this class than with the other, and faster to serialize, because it stores the
    objects as raw serializable data, but it also means it has worse operation support, because it needs to convert
    the raw data back.
    
    Attributes
    ----------
    _data : `dict` of (`str`, `Any`) items
        The raw data of the embed. It should not be accessed directly. There are several properties and methods to do
        operations on them.
    
    Examples
    --------
    Example of using local embed file:
    
    ```py
    # Imports
    from hata import Embed, ReuAsyncIO
    
    # Building the embed
    embed = Embed()
    embed.add_image('attachment://image.png')
    
    # Sending the message
    with (await ReuAsyncIO('some_file_path')) as file:
        await client.message_create(channel, embed=embed, file=('image.png', file))
    ```
    
    Note that you should use async io wrappers, but one which do not closes on `.close` either, but it resets
    itself instead, because if the request fails, the io would be closed and the request could not be done the
    second time.
    """
    __slots__ = ('_data',)
    
    def __init__(self, title=None, description=None, color=None, url=None, timestamp=None, type_='rich'):
        """
        Creates an embed instance. Accepts the base parameters of the embed.
        
        Parameters
        ----------
        title : `str`, Optional
            The title of the embed. Shows at the top with intense white characters.
        description : `str`, Optional
            The main content of the embed.
        color : ``Color`` or `int`, Optional
            The color code of the embed. Passing `0` means black, not like at the case of roles.
        url : `str`, Optional
            Url of the embed. If defined, the embed's `title` will show up as a hyper link pointing to the `url`.
        timestamp : `datetime`, optional
            Timestamp of the embed's content. Shows up next to the `footer` separated with a `'|'` character.
        type_ : `None` or `str`, Optional
            The type of the embed. Defaults to `'rich'`.
        """
        self._data = data = {}
        
        if (title is not None):
            if not isinstance(title, str):
                title = str(title)
            
            data['title'] = title
            
        if (description is not None):
            if not isinstance(description, str):
                description = str(description)
            
            data['description'] = description
        
        if (color is not None):
            data['color'] = color
            
        if (url is not None):
            data['url'] = url
            
        if (timestamp is not None):
            data['timestamp'] = datetime_to_timestamp(timestamp)
            
        if (type_ is not None):
            data['type'] = type_
    
    
    @copy_docs(EmbedBase.__len__)
    def __len__(self):
        data = self._data
        result = 0
        
        try:
            title = data['title']
        except KeyError:
            pass
        else:
            result += len(title)
        
        try:
            description = data['description']
        except KeyError:
            pass
        else:
            result += len(description)
        
        try:
            author_data = data['author']
        except KeyError:
            pass
        else:
            try:
                author_name = author_data['name']
            except KeyError:
                pass
            else:
                result += len(author_name)
        
        try:
            footer_data = data['footer']
        except KeyError:
            pass
        else:
            result += len(footer_data['text'])
        
        try:
            field_datas = data['fields']
        except KeyError:
            pass
        else:
            for field_data in field_datas:
                result += len(field_data['name'])
                result += len(field_data['value'])
        
        return result
    
    
    @copy_docs(EmbedBase.__bool__)
    def __bool__(self):
        data = self._data
        data_length = len(data) - ('type' in data)
        if data_length == 0:
            return False
        
        if data_length == 1:
            try:
                field_datas = data['fields']
            except KeyError:
                pass
            else:
                if not field_datas:
                    return False
        
        return True
    
    
    @property
    def contents(self):
        """
        Returns the embed's contents.
        
        The embeds contents are the following:
        - `.title`
        - `.description`
        - `.author.name`
        - `.footer.text`
        - `.fields[n].name`
        - `.fields[n].value`
        
        Returns
        -------
        contents : `list` of `str`
        """
        data = self._data
        result = []
        
        try:
            title = data['title']
        except KeyError:
            pass
        else:
            result.append(title)
        
        try:
            description = data['description']
        except KeyError:
            pass
        else:
            result.append(description)
        
        try:
            author_data = data['author']
        except KeyError:
            pass
        else:
            try:
                author_name = author_data['name']
            except KeyError:
                pass
            else:
                result.append(author_name)
        
        try:
            footer_data = data['footer']
        except KeyError:
            pass
        else:
            result.append(footer_data['text'])
        
        try:
            field_datas = data['fields']
        except KeyError:
            pass
        else:
            for field_data in field_datas:
                result.append(field_data['name'])
                result.append(field_data['value'])
        
        return result
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an embed from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Embed data received from Discord.
        
        Returns
        -------
        self : ``Embed``
        """
        self = object.__new__(cls)
        self._data = data
        return self
    
    
    def to_data(self):
        """
        Returns the embed's `._data`.
        
        This method is for compatibility with other embed-likes. When sending embed in message this method is called
        for getting it's data.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return self._data
    
    
    @copy_docs(EmbedBase.clear)
    def clear(self):
        data = self._data
        fields = data.get('fields', None)
        data.clear()
        if (fields is not None):
            fields.clear()
            data['fields'] = fields
    
    
    @copy_docs(EmbedBase.copy)
    def copy(self):
        new_data = {}
        
        for key, value in self._data.items():
            if value is None:
                continue
            
            if key in RICH_EMBED_FIELDS:
                if key == 'fields':
                    value = [field_data.copy() for field_data in value]
                
                else:
                    value = value.copy()
            
            new_data[key] = value
        
        new = object.__new__(type(self))
        new._data = new_data
        return new
    
    
    @copy_docs(EmbedBase.copy_with)
    def copy_with(self, **kwargs):
        data = self._data
        
        try:
            author = kwargs.pop('author')
        except KeyError:
            author_data = data.get('author', None)
            if (author_data is not None):
                author_data = author_data.copy()
        else:
            if (author is None):
                author_data = None
            else:
                author_data = author.to_data()
        
        try:
            color = kwargs.pop('color')
        except KeyError:
            color = data.get('color', None)
        
        try:
            description = kwargs.pop('description')
        except KeyError:
            description = data.get('description', None)
        
        try:
            fields = kwargs.pop('fields')
        except KeyError:
            field_datas = data.get('fields', None)
            if (field_datas is not None):
                field_datas = [field_data.copy() for field_data in field_datas]
        else:
            if (fields is None):
                field_datas = None
            else:
                field_datas = [field.to_data() for field in fields]
        
        try:
            footer = kwargs.pop('footer')
        except KeyError:
            footer_data = data.get('footer', None)
            if (footer_data is not None):
                footer_data = footer_data.copy()
        else:
            if (footer is None):
                footer_data = None
            else:
                footer_data = footer.to_data()
        
        try:
            image = kwargs.pop('image')
        except KeyError:
            image_data = data.get('image', None)
            if (image_data is not None):
                image_data = image_data.copy()
        else:
            if (image is None):
                image_data = None
            else:
                image_data = image.to_data()
        
        try:
            provider = kwargs.pop('provider')
        except KeyError:
            provider_data = data.get('provider', None)
            if (provider_data is not None):
                provider_data = provider_data.copy()
        else:
            if (provider is None):
                provider_data = None
            else:
                provider_data = provider.to_data()
        
        try:
            thumbnail = kwargs.pop('thumbnail')
        except KeyError:
            thumbnail_data = data.get('thumbnail', None)
            if (thumbnail_data is not None):
                thumbnail_data = thumbnail_data.copy()
        else:
            if (thumbnail is None):
                thumbnail_data = None
            else:
                thumbnail_data  = thumbnail.to_data()
        
        try:
            timestamp = kwargs.pop('timestamp')
        except KeyError:
            timestamp_data = data.get('timestamp', None)
        else:
            if (timestamp is None):
                timestamp_data = None
            else:
                timestamp_data = datetime_to_timestamp(timestamp)
        
        try:
            title = kwargs.pop('title')
        except KeyError:
            title = data.get('title', None)
        
        try:
            type_ = kwargs.pop('type')
        except KeyError:
            type_ = data.get('type_', None)
        
        try:
            url = kwargs.pop('url')
        except KeyError:
            url = data.get('url', None)
        
        try:
            video = kwargs.pop('video')
        except KeyError:
            video_data = data.get('video', None)
            if (video_data is not None):
                video_data = video_data.copy()
        else:
            if (video is None):
                video_data = None
            else:
                video_data = video.to_data()
        
        if kwargs:
            raise TypeError(f'Unused or unsettable attributes: `{kwargs}`')
        
        
        new_data = {}
        
        if (author_data is not None):
            new_data['author'] = author_data
        
        if (color is not None):
            new_data['color'] = color
        
        if (description is not None):
            if not isinstance(description, str):
                description = str(description)
            
            new_data['description'] = description
        
        if (field_datas is not None):
            new_data['fields'] = field_datas
        
        if (footer_data is not None):
            new_data['footer'] = footer_data
        
        if (image_data is not None):
            new_data['image'] = image_data
        
        if (provider_data is not None):
            new_data['provider'] = provider_data
        
        if (thumbnail_data is not None):
            new_data['thumbnail'] = thumbnail_data
        
        if (timestamp_data is not None):
            new_data['timestamp'] = timestamp_data
        
        if (title is not None):
            if not isinstance(title, str):
                title = str(title)
            
            new_data['title'] = title
        
        if (type_ is not None):
            new_data['type_'] = type_
        
        if (url is not None):
            new_data['url'] = url
        
        if (video_data is not None):
            new_data['video'] = video_data
        
        new = object.__new__(type(self))
        new._data = new_data
        return new
    
    # Properties
    
    # `.author`
    
    @property
    def author(self):
        """
        A get-set-del property for accessing the embed's author.
        
        Accepts and returns `None` or an ``EmbedAuthor`` object.
        """
        try:
            author_data = self._data['author']
        except KeyError:
            return None
        
        return EmbedAuthor.from_data(author_data)
    
    @author.setter
    def author(self, value):
        self._data['author'] = value.to_data()
    
    @author.deleter
    def author(self):
        try:
            del self._data['author']
        except KeyError:
            pass
    
    # `.color`
    
    @property
    def color(self):
        """
        A get-set-del property for accessing the embed's color.
        
        Accepts and returns `None` or a ``Color`` (/ `int`) object.
        """
        return self._data.get('color', None)
    
    @color.setter
    def color(self, value):
        self._data['color'] = value
    
    @color.deleter
    def color(self):
        try:
            del self._data['color']
        except KeyError:
            pass
    
    # `.description`
    
    @property
    def description(self):
        """
        A get-set-del property for accessing the embed's description.
        
        Accepts and returns `None` or a `str` instance.
        """
        return self._data.get('description', None)
    
    @description.setter
    def description(self, value):
        if value is None:
            try:
                del self._data['description']
            except KeyError:
                pass
        else:
            if not isinstance(value, str):
                value = str(value)
            
            self._data['description'] = value
    
    @description.deleter
    def description(self):
        try:
            del self._data['description']
        except KeyError:
            pass
    
    # `.fields`
    
    @property
    def fields(self):
        try:
            field_datas = self._data['fields']
        except KeyError:
            self._data['fields'] = field_datas = []
        
        return _EmbedFieldsProxy(field_datas)
    
    @fields.setter
    def fields(self, value):
        """
        A get-set-del property for accessing the embed's fields.
        
        Accepts an `iterable` of ``EmbedField``objects. Meanwhile returns an ``_EmbedFieldsProxy`` instance,
        through what the respective embed's fields can be modified directly.
        """
        data = self._data
        try:
            fields_data = data['fields']
        except KeyError:
            fields_data = data['fields'] = []
        
        if type(value) is _EmbedFieldsProxy:
            new_fields_data = value._data
        else:
            new_fields_data = list(field.to_data() for field in value)
        
        fields_data.clear()
        fields_data.extend(new_fields_data)
    
    @fields.deleter
    def fields(self):
        try:
            field_datas = self._data['fields']
        except KeyError:
            pass
        else:
            field_datas.clear()
    
    # `.footer`
    
    @property
    def footer(self):
        """
        A get-set-del property for accessing the embed's footer.
        
        Accepts and returns `None` or an ``EmbedFooter`` object.
        """
        try:
            footer_data = self._data['footer']
        except KeyError:
            return None
        
        return EmbedFooter.from_data(footer_data)
    
    @footer.setter
    def footer(self, value):
        self._data['footer'] = value.to_data()
    
    @footer.deleter
    def footer(self):
        try:
            del self._data['footer']
        except KeyError:
            pass
    
    # `.image`
    
    @property
    def image(self):
        """
        A get-set-del property for accessing the embed's image.
        
        Accepts and returns `None` or an ``EmbedImage`` object.
        """
        try:
            image_data = self._data['image']
        except KeyError:
            return None
        
        return EmbedImage.from_data(image_data)
    
    @image.setter
    def image(self, value):
        self._data['image'] = value.to_data()
    
    @image.deleter
    def image(self):
        try:
            del self._data['image']
        except KeyError:
            pass
        
    # `.provider`
    
    @property
    def provider(self):
        """
        A get-del property for accessing the embed's provider.
        
        Returns `None` or an ``EmbedProvider`` object.
        
        Embed providers cannot be set, they are receive only.
        """
        try:
            provider_data = self._data['provider']
        except KeyError:
            return None
        
        return EmbedProvider.from_data(provider_data)
    
    @provider.deleter
    def provider(self):
        try:
            del self._data['provider']
        except KeyError:
            pass
    
    # `.thumbnail`
    
    @property
    def thumbnail(self):
        """
        A get-set-del property for accessing the embed's thumbnail.
        
        Accepts and returns `None` or an ``EmbedThumbnail`` object.
        """
        try:
            thumbnail_data = self._data['thumbnail']
        except KeyError:
            return None
        
        return EmbedThumbnail.from_data(thumbnail_data)
    
    @thumbnail.setter
    def thumbnail(self, value):
        self._data['thumbnail'] = value.to_data()
    
    @thumbnail.deleter
    def thumbnail(self):
        try:
            self._data['thumbnail']
        except KeyError:
            pass
    
    # `.timestamp`
    
    @property
    def timestamp(self):
        """
        A get-set-del property for accessing the embed's timestamp.
        
        Accepts and returns `None` or a `datetime` object.
        """
        try:
            timestamp_value = self._data['timestamp']
        except KeyError:
            return None
        
        return timestamp_to_datetime(timestamp_value)
    
    @timestamp.setter
    def timestamp(self, value):
        self._data['timestamp'] = datetime_to_timestamp(value)
    
    @timestamp.deleter
    def timestamp(self):
        try:
            del self._data['timestamp']
        except KeyError:
            pass
    
    # `.title`
    
    @property
    def title(self):
        """
        A get-set-del property for accessing the embed's title.
        
        Accepts and returns `None` or a `str` instance.
        """
        return self._data.get('title', None)
    
    @title.setter
    def title(self, value):
        if value is None:
            try:
                del self._data['title']
            except KeyError:
                pass
        else:
            if not isinstance(value, str):
                value = str(value)
            
            self._data['title'] = value
    
    @title.deleter
    def title(self):
        try:
            del self._data['title']
        except KeyError:
            pass
    
    # `.type`
    
    @property
    def type(self):
        """
        A get-set-del property for accessing the embed's type.
        
        Accepts and returns `None` or a `str` instance.
        """
        return self._data.get('type', None)
    
    @type.setter
    def type(self, value):
        self._data['type'] = value
    
    @type.deleter
    def type(self):
        try:
            del self._data['type']
        except KeyError:
            pass
    
    # `.url`
    
    @property
    def url(self):
        """
        A get-set-del property for accessing the embed's url.
        
        Accepts and returns `None` or a `str` instance.
        """
        return self._data.get('url', None)
    
    @url.setter
    def url(self, value):
        self._data['url'] = value
    
    @url.deleter
    def url(self):
        try:
            del self._data['url']
        except KeyError:
            pass
    
    # `.video`
    
    @property
    def video(self):
        """
        A get-del property for accessing the embed's video.
        
        Returns `None` or an ``EmbedVideo`` object.
        
        Embed videos cannot be set, they are receive only.
        """
        try:
            video_data = self._data['video']
        except KeyError:
            return None
        
        return EmbedVideo.from_data(video_data)
    
    @video.deleter
    def video(self):
        try:
            del self._data['video']
        except KeyError:
            pass
    
    # Methods
    
    # author
    
    def add_author(self, icon_url=None, name=None, url=None):
        """
        Adds an ``EmbedAuthor`` to the embed with the given parameters.
        
        Parameters
        ----------
        icon_url : `str`, Optional
            An url of the author's icon. Can be http(s) or attachment.
        name : `str`, Optional
            The name of the author.
        url : `str`, Optional
            The url of the author.
        
        Returns
        -------
        self : ``Embed``
        """
        if not isinstance(name, str):
            name = str(name)
        
        author_data = {}
        
        if (name is not None):
            author_data['name'] = name
        
        if (url is not None):
            author_data['url'] = url
        
        if (icon_url is not None):
            author_data['icon_url'] = icon_url
        
        self._data['author'] = author_data
        return self
    
    # fields

    def add_field(self, name, value, inline=False):
        """
        Adds an ``EmbedField`` to the end of the embed's fields.
        
        Parameters
        ----------
        name : `str`
            The name of the field.
        value : `str`
            The value of the field.
        inline : `bool`, Optional
            Whether this field should display inline.
        
        Returns
        -------
        self : ``Embed``
        """
        if not isinstance(name, str):
            name = str(name)
        
        if not isinstance(value, str):
            value = str(value)
        
        field_data = {
            'name': name,
            'value': value,
        }
        
        if inline:
            field_data['inline'] = inline
        
        try:
            field_datas = self._data['fields']
        except KeyError:
            self._data['fields'] = [field_data]
        else:
            field_datas.append(field_data)
        
        return self
    
    
    def insert_field(self, index, name, value, inline=False):
        """
        Inserts an ``EmbedField`` to the embed's fields at the specified `index`.
        
        Parameters
        ----------
        index : `int`
            The index before the field should be inserted.
        name : `str`
            The name of the field.
        value : `str`
            The value of the field.
        inline : `bool`, Optional
            Whether this field should display inline.
        """
        if not isinstance(name, str):
            name = str(name)
        
        if not isinstance(value, str):
            value = str(value)
        
        field_data = {
            'name': name,
            'value': value,
        }
        
        if inline:
            field_data['inline'] = inline
        
        try:
            field_datas = self._data['fields']
        except KeyError:
            self._data['fields'] = [field_data]
        else:
            field_datas.insert(index, field_data)
    
    
    def get_field(self, index):
        """
        Returns the embed's field at the given `index`.
        
        Parameters
        ----------
        index : `int`
            The index of the field.
        
        Returns
        -------
        embed_field : ``EmbedField``
        
        Raises
        ------
        IndexError
            Index out of the fields' range.
        """
        try:
            field_datas = self._data['fields']
        except KeyError:
            raise IndexError('Index out of the fields\' range.') from None
        
        try:
            field_data = field_datas[index]
        except IndexError as err:
            err.args = ('Index out of the fields\' range.', )
            raise
        
        return EmbedField.from_data(field_data)
    
    
    def append_field(self, field):
        """
        Appends the embed's fields with the given field.
        
        Parameters
        ----------
        field : ``EmbedField``
            The field to append the embed's field with.
        """
        field_data = field.to_data()
        
        try:
            field_datas = self._data['fields']
        except KeyError:
            self._data['fields'] = [field_data]
        else:
            field_datas.append(field_data)
    
    
    def set_field(self, index, field):
        """
        Sets the given ``EmbedField`` at the given `index` to the fields of the embed.
        
        Parameters
        ----------
        index : `int`
            The index of the field to set.
        field : ``EmbedField``
            The embed field to set at the given `index`.
        
        Raises
        ------
        IndexError
            Index out of the fields' range.
        """
        try:
            fields = self._data['fields']
        except KeyError:
            raise IndexError('Index out of the fields\' range.') from None
        
        field_data = field.to_data()
        
        try:
            fields[index] = field_data
        except IndexError as err:
            err.args = ('Index out of the fields\' range.',)
            raise
    
    
    def del_field(self, index):
        """
        Removes the field of the embed on the given index.
        
        Parameters
        ----------
        index : `int`
            The index of the field to remove.
        
        Raises
        ------
        IndexError
            Index out of the fields' range.
        """
        try:
            fields = self._data['fields']
        except KeyError:
            raise IndexError('Index out of the fields\' range.') from None
        
        try:
            del fields[index]
        except IndexError as err:
            err.args = ('Index out of the fields\' range.', )
            raise
    
    remove_field = del_field
    
    # footer
    
    def add_footer(self, text, icon_url=None):
        """
        Adds an ``EmbedFooter`` to the embed with the given parameters.
        
        Parameters
        ----------
        text : `str`
            The footer's text.
        icon_url : `str`, Optional
            An url of the footer's icon. Can be http(s) or attachment.
        
        Returns
        -------
        self : ``Embed``
        """
        if not isinstance(text, str):
            text = str(text)
        
        footer_data = {
            'text': text,
        }
        
        if (icon_url is not None):
            footer_data['icon_url'] = icon_url
        
        self._data['footer'] = footer_data
        return self
    
    # image
    
    def add_image(self, url):
        """
        Adds an ``EmbedImage`` to the embed with the given `url`.
        
        Parameters
        ----------
        url : `str`
            The url of the image. Can be http(s) or attachment.
        
        Returns
        -------
        self : ``Embed``
        """
        image_data = {
            'url': url,
        }
        
        self._data['image'] = image_data
        return self
    
    # thumbnail
    
    def add_thumbnail(self, url):
        """
        Adds an ``EmbedThumbnail`` to the embed with the given `url`.
        
        Parameters
        ----------
        url : `str`
            The url of the thumbnail. Can be http(s) or attachment.
        
        Returns
        -------
        self : ``Embed``
        """
        thumbnail_data = {
            'url': url,
        }
        
        self._data['thumbnail'] = thumbnail_data
        return self


class _EmbedFieldsProxy:
    """
    A reflection of an `Embed` object's fields. The instances of this type allow direct modifications of their
    respective embed's fields.
    
    Attributes
    ----------
    _data : `list` of (`dict` of (`str`, `Any`) items)
        Raw data containing the respective embed's fields.
    """
    __slots__ = ('_data',)
    def __init__(self, data):
        """
        Creates a ``_EmbedFieldsProxy`` object.
        
        Parameters
        ----------
        data : `list` of (`dict` of (`str`, `Any`) items)
            Raw data containing the respective embed's fields.
        """
        self._data = data
        
    def clear(self):
        """
        Removes all of the respective embed's fields.
        """
        self._data.clear()
    
    def __len__(self):
        """Returns how much fields the respective embed has."""
        return len(self._data)
    
    def __repr__(self):
        """Returns the representation of the object."""
        return f'<{self.__class__.__name__} length={len(self._data)}>'
    
    def __getitem__(self, index):
        """Returns the embed field on the given index."""
        return EmbedField.from_data(self._data[index])
    
    def __setitem__(self, index, field):
        """Sets the given embed field object on the given index."""
        self._data[index] = field.to_data()
    
    def __delitem__(self, index):
        """Deletes the field on the given index"""
        del self._data[index]
    
    def append(self, field):
        """
        Appends the respective embed's fields with the given field.
        
        Parameters
        ----------
        field : ``EmbedField``
            The field to append the embed's field with.
        """
        self._data.append(field.to_data())
    
    def insert(self, index, field):
        """
        Inserts the given `field` to the respective embed's fields at the specified `index`.
        
        Parameters
        ----------
        index : `int`
            The index before the field should be inserted.
        field : ``EmbedField``
            The field to insert.
        """
        self._data.insert(index,field.to_data())
    
    def add_field(self, name, value, inline=False):
        """
        Adds an ``EmbedField`` to the end of the respective embed's fields.
        
        Parameters
        ----------
        name : `str`
            The name of the field.
        value : `str`
            The value of the field.
        inline : `bool`, Optional
            Whether this field should display inline.
        """
        if not isinstance(name, str):
            name = str(name)
        
        if not isinstance(value, str):
            value = str(value)
        
        field_data = {
            'name': name,
            'value': value,
        }
        
        if inline:
            field_data['inline'] = inline
        
        self._data.append(field_data)
    
    
    def insert_field(self, index, name, value, inline=False):
        """
        Inserts an ``EmbedField`` to the respective embed's fields at the specified `index`.
        
        Parameters
        ----------
        index : `int`
            The index before the field should be inserted.
        name : `str`
            The name of the field.
        value : `str`
            The value of the field.
        inline : `bool`, Optional
            Whether this field should display inline.
        """
        if not isinstance(name, str):
            name = str(name)
        
        if not isinstance(value, str):
            value = str(value)
        
        field_data = {
            'name' : name,
            'value' : value,
        }
        
        if inline:
            field_data['inline'] = inline
        
        self._data.insert(index, field_data)
    
    
    def __iter__(self):
        """
        Iterates over the respective embed's fields.
        
        This method is a generator.
        """
        for field_data in self._data:
            yield EmbedField.from_data(field_data)
    
    
    def __reversed__(self):
        """
        Reverse iterates over the respective embed's fields.
        
        This method is a generator.
        """
        for field_data in reversed(self._data):
            yield EmbedField.from_data(field_data)



# -*- coding: utf-8 -*-
__all__ = ('Formdata', )

from io import IOBase
from urllib.parse import urlencode
from json import dumps as dump_to_json

from .utils import multidict

from .headers import CONTENT_TYPE, CONTENT_TRANSFER_ENCODING, CONTENT_LENGTH
from .multipart import MultipartWriter, create_payload, BytesPayload


class Formdata:
    """
    Helper class for `multipart/form-data` and `application/x-www-form-urlencoded` body generation.
    
    Attributes
    ----------
    fields : `list` of `tuple` (`multidict` of (`str`, `str`) items, `multidict` of (`str`, `str`) items, `Any`)
        The fields of the formdata. Each element is a tuple, which contains the following elements:
        - `type_options` : `multidict` of (`str`, `str`) items;
            Additional information used by the created ``PayloadBase`` instance when the field is generated.
        - `headers` : `multidict` of (`str`, `str`) items;
            The field specific headers.
        - value : `Any`;
            The field's value.
    
    is_multipart : `bool`
        Whether the formdata is `multipart/form-data` and not `application/x-www-form-urlencoded` type.
    
    quote_fields : `bool`
        Whether type option field values should be quoted.
    """
    __slots__ = ('fields', 'is_multipart', 'quote_fields', )
    
    def __init__(self, quote_fields=True):
        """
        Creates a new a ``Formdata``.
        
        Parameters
        ----------
        quote_fields : `bool`, Optional
            Whether type option field values should be quoted. Defaults to `True`.
        """
        self.fields = []
        self.is_multipart = False
        self.quote_fields = quote_fields
    
    @classmethod
    def from_fields(cls, fields):
        """
        Creates a new ``Formdata`` instance from the given fields.
        
        Parameters
        ----------
        fields : `dict` of (`str`, `Any`) items, `list` of `tuple` (`str`, `Any`), `IOBase` instance
            The fields to convert to ``Formdata``.
        
        Returns
        -------
        self : ``Formdata``
            The created formdata instance.
        
        Raises
        ------
        TypeError
            Received unhandleable field.
        """
        if isinstance(fields, dict):
            fields = list(fields.items())
        elif isinstance(fields, (list, tuple)):
            fields = list(fields)
        else:
            fields = [fields]
        
        self = cls()
        
        while fields:
            field = fields.pop(0)
            fiend_type = field.__class__
            if issubclass(fiend_type, IOBase):
                self.add_field(getattr(field, 'name', 'unknown'), field)
            elif issubclass(fiend_type, (tuple, list)) and len(field) == 2:
                self.add_field(*field)
            else:
                raise TypeError(f'`Formdata.from_fields` got unhandleable field: {fiend_type.__name__}; {field!r}.')
        
        return self
        
    def add_field(self, name, value, content_type=None, filename=None, transfer_encoding=None):
        """
        Adds a field to the formdata.
        
        Parameters
        ----------
        name : `str`
            The field's name.
        value : `Any`
            The field's value.
        content_type : `None` or `str`, Optional
            The field's content type. Defaults to `None`.
        filename : `None` or `str`, Optional
            The field's name. If not given or given as `None` (so by the default), and the given `value` is an `IOBase`
            instance then tries to use that's name.
        transfer_encoding : `None` or `str`, Optional
            The field's transfer encoding. Defaults to `None`.
        
        Raises
        ------
        AssertionError
            - If `content_type` was not given as `None`, neither as `str` instance.
            - If `filename` was not given as `None`, neither as `str` instance.
            - If `transfer_encoding` was not given as `None`, neither as `str` instance.
        """
        if isinstance(value, IOBase):
            self.is_multipart = True
        elif isinstance(value, (bytes, bytearray, memoryview)):
            if (filename is None) and (transfer_encoding is None):
                filename = name
        
        type_options = multidict()
        type_options['name'] = name
        
        if (filename is not None):
            if __debug__:
                if not isinstance(filename, str):
                    raise AssertionError(f'`filename` can be given as `None` or `str` instance, got '
                        f'{filename.__class__.__name__}.')
        
        if (filename is None) and isinstance(value, IOBase):
            filename = getattr(value, 'name', name)
        
        if (filename is not None):
            type_options['filename'] = filename
            self.is_multipart = True
        
        headers = {}
        if (content_type is not None):
            if __debug__:
                if not isinstance(content_type, str):
                    raise AssertionError('`content_type` can be `None` or `str` instance, got '
                        f'{content_type.__class__.__name__}.')
            
            headers[CONTENT_TYPE] = content_type
            self.is_multipart = True
        
        if (transfer_encoding is not None):
            if __debug__:
                if not isinstance(transfer_encoding, str):
                    raise AssertionError('`transfer_encoding` can be `None` or `str` instance, got: '
                        f'{transfer_encoding.__class__.__name__}.')
            
            headers[CONTENT_TRANSFER_ENCODING] = transfer_encoding
            self.is_multipart = True
        
        self.fields.append((type_options, headers, value))
    
    def _gen_form_urlencoded(self, encoding):
        """
        Generates `application/x-www-form-urlencoded` payload from the ``Formdata``'s fields.
        
        Parameters
        ----------
        encoding : `str`
            The encoding to use to encode the formdata's fields.
        
        Returns
        -------
        payload : ``BytesPayload``
            The generated payload.
        """
        data = []
        for type_options, header, value in self.fields:
            data.append((type_options['name'], value))
        
        if encoding == 'utf-8':
            content_type = 'application/x-www-form-urlencoded'
        else:
            content_type = f'application/x-www-form-urlencoded; charset={encoding}'
        
        return BytesPayload(urlencode(data, doseq=True, encoding=encoding).encode(), {'content_type': content_type})
    
    def _gen_form_data(self, encoding):
        """
        Generates `multipart/form-data` payload from the ``Formdata``'s fields.
        
        Parameters
        ----------
        encoding : `str`
            The encoding to use to encode the formdata's fields.
        
        Returns
        -------
        payload : ``MultipartWriter``
            The generated payload.
        
        Raises
        ------
        TypeError
            Cannot serialize a field.
        RuntimeError
            - If a field's content has unknown content-encoding.
            - If a field's content has unknown content-transfer-encoding.
        """
        writer = MultipartWriter('form-data')
        for type_options, headers, value in self.fields:
            try:
                payload_kwargs = {
                    'headers': headers,
                    'encoding': encoding,
                }
                
                try:
                    content_type = headers[CONTENT_TYPE]
                except KeyError:
                    pass
                else:
                    payload_kwargs['content_type'] = content_type
                    
                    if type_options:
                        payload_kwargs.update(type_options.kwargs())
                
                part = create_payload(value, payload_kwargs)
            
            except BaseException as err:
                raise TypeError(f'Can not serialize value type: {value.__class__.__name__}, headers: {headers!r}, '
                    f'value: {value!r}.') from err
            
            if type_options:
                part.set_content_disposition('form-data', type_options.kwargs(), quote_fields=self.quote_fields)
                part.headers.pop_all(CONTENT_LENGTH, None)
            
            writer.append_payload(part)
        
        return writer
    
    def __call__(self, encoding='utf-8'):
        """
        Gets the payload of the ``Formdata``-
        
        Parameters
        ----------
        encoding : `str`
            The encoding to use to encode the formdata's fields.
        
        Returns
        -------
        payload : ``BytesPayload`` or ``MultipartWriter``
            The generated payload.
        
        Raises
        ------
        TypeError
            Cannot serialize a field.
        """
        if self.is_multipart:
            return self._gen_form_data(encoding)
        else:
            return self._gen_form_urlencoded(encoding)
    
    def add_json(self, data):
        """
        Shortcut to add a json field to the ``Formdata``.
        
        Parameters
        ----------
        data : `Any`
            Json serializable content.
        """
        if data:
            type_options = multidict()
            type_options['name'] = 'data_json'
            data = dump_to_json(data, separators=(',', ':'), ensure_ascii=True)
            self.fields.append((type_options, multidict(), data))
    
    def __repr__(self):
        """Returns the representation of the formdata."""
        result = ['<', self.__class__.__name__, ' [']
        
        fields = self.fields
        limit = len(fields)
        if limit:
            index = 0
            while True:
                type_options, headers, value = fields[index]
                result.append('(')
                result.append(repr(type_options))
                result.append(', ')
                result.append(repr(headers))
                result.append(', ')
                result.append(repr(value))
                result.append(')')
                
                index += 1
                if index == limit:
                    break
                    
                result.append(', ')
                continue
        
        result.append(']>')
        return ''.join(result)
    
    __str__ = __repr__



__all__ = ('create_partial_emoji_data', 'create_partial_emoji_from_data', 'create_unicode_emoji',
    'parse_custom_emojis', 'parse_emoji', 'parse_reaction')

import warnings

from ...backend.export import export

from ..utils import EMOJI_RP, REACTION_RP

from .emoji import UNICODE_TO_EMOJI, Emoji


@export
def create_partial_emoji_from_data(data):
    """
    Creates an emoji from partial emoji data sent by Discord.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        Partial emoji data.
    
    Returns
    -------
    emoji : ``Emoji``
    """
    try:
        emoji_name = data['name']
    except KeyError:
        emoji_name = data['emoji_name']
        emoji_id = data.get('emoji_id', None)
        emoji_animated = data.get('emoji_animated', False)
    else:
        emoji_id = data.get('id', None)
        emoji_animated = data.get('animated', False)
    
    if emoji_id is None:
        try:
            emoji = UNICODE_TO_EMOJI[emoji_name]
        except KeyError:
            warnings.warn(
                f'Undefined emoji : {emoji_name.encode()!r}\nPlease open an issue with this message.',
                RuntimeWarning,
            )
            emoji = Emoji._create_unicode('', emoji_name, ())
    
    else:
        # name can change
        if emoji_name is None:
            emoji_name = ''
            
        emoji = Emoji._create_partial(int(emoji_id), emoji_name, emoji_animated)
    
    return emoji


def create_partial_emoji_data(emoji):
    """
    Creates partial emoji data form the given emoji.
    
    Parameters
    ----------
    emoji : ``Emoji``
        The emoji to serialize.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
        The serialized emoji data.
    """
    emoji_data = {}
    unicode = emoji.unicode
    if unicode is None:
        emoji_data['id'] = emoji.id
        emoji_data['name'] = emoji.name
        
        if emoji.animated:
            emoji_data['animated'] = True
    else:
        emoji_data['name'] = unicode
    
    return emoji_data


def parse_emoji(text):
    """
    Tries to parse out an ``Emoji`` from the inputted text. This emoji can be custom and unicode emoji as well.
    
    If the parsing yields a custom emoji what is not loaded, the function will return an `untrusted` partial emoji,
    what means it wont be stored at `EMOJIS`. If the parsing fails the function returns `None`.
    
    Returns
    -------
    emoji : `None` or ``Emoji``
    """
    parsed = EMOJI_RP.fullmatch(text)
    if parsed is None:
        emoji = UNICODE_TO_EMOJI.get(text, None)
    else:
        animated, name, emoji_id = parsed.groups()
        animated = (animated is not None)
        emoji_id = int(emoji_id)
        emoji = Emoji._create_partial(emoji_id, name, animated)
    
    return emoji


def parse_custom_emojis(text):
    """
    Parses out every custom emoji from the given text.
    
    Parameters
    ----------
    text : `str`
        Text, what might contain custom emojis.
    
    Returns
    -------
    emojis : `set` of ``Emoji``
    """
    emojis = set()
    for groups in EMOJI_RP.findall(text):
        animated, name, emoji_id = groups
        animated = (animated is not None)
        emoji_id = int(emoji_id)
        emoji = Emoji._create_partial(emoji_id, name, animated)
        emojis.add(emoji)
    
    return emojis


def parse_reaction(text):
    """
    Parses out an emoji from the given reaction string.
    
    Parameters
    ----------
    text : `str`
        Reaction string.
    
    Returns
    -------
    emoji : `None` or ``Emoji``
    """
    try:
        emoji = UNICODE_TO_EMOJI[text]
    except KeyError:
        parsed = REACTION_RP.fullmatch(text)
        if parsed is None:
            emoji = None
        else:
            name, emoji_id = parsed.parsed()
            emoji_id = int(emoji_id)
            emoji = Emoji._create_partial(emoji_id, name, False)
    
    return emoji


@export
def create_unicode_emoji(unicode):
    """
    Creates an emoji from the given unicode value.
    
    Parameters
    ----------
    unicode : `str`
        Unicode value.
    
    Returns
    -------
    emoji : ``Emoji``
    """
    try:
        unicode_emoji = UNICODE_TO_EMOJI[unicode]
    except KeyError:
        warnings.warn(
            f'Undefined emoji : {unicode.encode()!r}\nPlease open an issue with this message.',
            RuntimeWarning,
        )
        unicode_emoji = Emoji._create_unicode('', unicode, ())
    
    return unicode_emoji



__all__ = ('EventBase', )

class EventBase:
    """
    Base class for events.
    """
    __slots__ = ()
    
    def __new__(cls, *args, **kwargs):
        return None
    
    def __repr__(self):
        """Returns the event's representation."""
        return f'<{self.__class__.__name__}>'
    
    def __len__(self):
        """Helper for unpacking if needed."""
        return 0
    
    def __iter__(self):
        """
        Unpacks the event.
        
        This method is a generator.
        """
        return
        yield # This is intentional. Python stuff... Do not ask, just accept.



from ..extension_loader.snapshot import SNAPSHOT_TAKERS
from ..extension_loader.extension_loader import EXTENSION_LOADER

from .application_command import SYNC_ID_NON_GLOBAL
from .slasher import Slasher
from .utils import RUNTIME_SYNC_HOOKS

def take_slasher_snapshot(client):
    """
    Collects all the command changes from the client's slash command processor.
    
    Parameters
    ----------
    client : ``Client``
        The client, who will be snapshotted.
    
    Returns
    -------
    collected : `None` or `tuple` of (`dict` of (`int`, `list` of `tuple` \
            (`bool`, ``SlasherApplicationCommand``)) items, `None` or `set` of ``ComponentCommand``)
        The collected commands of the slasher.
    """
    slasher = getattr(client, 'slasher', None)
    if (slasher is None) or (not isinstance(slasher, Slasher)):
        collected = None
    
    else:
        command_states = slasher._command_states
        
        collected_application_commands = None
        for guild_id, command_state in command_states.items():
            if guild_id == SYNC_ID_NON_GLOBAL:
                active_commands = command_state._active
                if (active_commands is None):
                    continue
                
                command_changes = [(True, command) for command in active_commands]
                
            else:
                changes = command_state._changes
                if changes is None:
                    continue
                
                command_changes = [tuple(change) for change in changes]
            
            if collected_application_commands is None:
                collected_application_commands = {}
            
            collected_application_commands[guild_id] = command_changes
        
        collected_component_commands = slasher._component_commands
        if collected_component_commands:
            collected_component_commands = collected_component_commands.copy()
        else:
            collected_component_commands = None
        
        if (collected_application_commands is None) and (collected_component_commands is None):
            collected = None
        else:
            collected = (collected_application_commands, collected_component_commands)
    
    return collected

def calculate_slasher_snapshot_difference(client, snapshot_old, snapshot_new):
    """
    Calculates the difference between two slasher snapshots
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    snapshot_old :  `None` or `tuple` of (`dict` of (`int`, `list` of `tuple` \
            (`bool`, ``SlasherApplicationCommand``)) items, `None` or `set` of ``ComponentCommand``)
        An old snapshot taken.
    snapshot_new :  `None` or `tuple` of (`dict` of (`int`, `list` of `tuple` \
            (`bool`, ``SlasherApplicationCommand``)) items, `None` or `set` of ``ComponentCommand``)
        A new snapshot.
    
    Returns
    -------
    snapshot_difference : `None` or `tuple` (`tuple` (`set` of ``SlasherApplicationCommand``, `set` of \
            ``SlasherApplicationCommand``), `tuple` (`None` or `set` of ``ComponentCommand``, `None` or \
            `set` of ``ComponentCommand``)
        The difference between the two snapshots.
    """
    if (snapshot_old is None) and (snapshot_new is None):
        return None
    
    
    if snapshot_old is None:
        application_command_snapshot_old = None
        component_command_snapshot_old = None
    else:
        application_command_snapshot_old, component_command_snapshot_old = snapshot_old
    
    if snapshot_new is None:
        application_command_snapshot_new = None
        component_command_snapshot_new = None
    else:
        application_command_snapshot_new, component_command_snapshot_new = snapshot_new
        
    
    if (application_command_snapshot_old is not None) or (application_command_snapshot_new is not None):
        added_application_commands = []
        removed_application_commands = []
        
        guild_ids = set()
        if (application_command_snapshot_old is not None):
            guild_ids.update(application_command_snapshot_old.keys())
        
        if (application_command_snapshot_new is not None):
            guild_ids.update(application_command_snapshot_new.keys())
        
        for guild_id in guild_ids:
            local_added_application_commands = []
            local_removed_application_commands = []
            
            if (application_command_snapshot_new is not None):
                try:
                    new_changes = application_command_snapshot_new[guild_id]
                except KeyError:
                    pass
                else:
                    for added, command in new_changes:
                        if added:
                            local_added_application_commands.append(command)
                        else:
                            local_removed_application_commands.remove(command)
            
            if (application_command_snapshot_old is not None):
                try:
                    old_changes = application_command_snapshot_old[guild_id]
                except KeyError:
                    pass
                else:
                    for added, command in old_changes:
                        if added:
                            try:
                                local_added_application_commands.remove(command)
                            except ValueError:
                                local_removed_application_commands.append(command)
                        else:
                            try:
                                local_removed_application_commands.remove(command)
                            except ValueError:
                                local_added_application_commands.append(command)
            
            added_application_commands.extend(local_added_application_commands)
            removed_application_commands.extend(local_removed_application_commands)
        
        if (not added_application_commands):
            added_application_commands = None
        
        if (not removed_application_commands):
            removed_application_commands = None
    
        if (added_application_commands is None) and (removed_application_commands is None):
            application_command_difference = None
        else:
            if client.running and client.application.id:
                slasher = getattr(client, 'slasher', None)
                if (slasher is not None):
                    slasher.sync()
            
            application_command_difference = added_application_commands, removed_application_commands
    else:
        application_command_difference = None
    
    if (component_command_snapshot_old is None) or (component_command_snapshot_new is None):
        removed_component_commands = component_command_snapshot_old
        added_component_commands = component_command_snapshot_new
    else:
        removed_component_commands = component_command_snapshot_old-component_command_snapshot_new
        added_component_commands = component_command_snapshot_new-component_command_snapshot_old
        
        if (not removed_component_commands):
            removed_component_commands = None
        
        if (not added_component_commands):
            added_component_commands = None
    
    if (added_component_commands is None) and (removed_component_commands is None):
        component_command_difference = None
    else:
        component_command_difference = (removed_component_commands, added_component_commands)
    
    if (application_command_difference is None) and (component_command_difference is None):
        snapshot_difference = None
    else:
        snapshot_difference = (application_command_difference, component_command_difference)
    
    return snapshot_difference

def revert_slasher_snapshot(client, snapshot_difference):
    """
    Reverts a snapshot taken from a slasher.
    
    Parameters
    ----------
    client : ``Client``
        The respective client instance.
    snapshot_difference : `None` or `tuple` (`tuple` (`set` of ``SlasherApplicationCommand``, `set` of \
            ``SlasherApplicationCommand``), `tuple` (`None` or `set` of ``ComponentCommand``, `None` or \
            `set` of ``ComponentCommand``)
        The taken snapshot.
    """
    slasher = getattr(client, 'slasher', None)
    if (slasher is None) or (not isinstance(slasher, Slasher)):
        return
    
    if (snapshot_difference is not None):
        application_command_difference, component_command_difference = snapshot_difference
        
        if (application_command_difference is not None):
            added_application_commands, removed_application_commands = application_command_difference
            if (added_application_commands is not None):
                for application_command in added_application_commands:
                    slasher._remove_application_command(application_command)
            
            if (removed_application_commands is not None):
                for application_command in removed_application_commands:
                    slasher._add_application_command(application_command)
            
            if client.running and client.application.id:
                slasher.sync()
        
        if (component_command_difference is not None):
            added_component_commands, removed_component_commands = component_command_difference
            
            if (added_component_commands is not None):
                for component_command in added_component_commands:
                    slasher._remove_component_command(component_command)

            if (removed_component_commands is not None):
                for component_command in removed_component_commands:
                    slasher._add_component_command(component_command)

SNAPSHOT_TAKERS['client.slasher'] = (
    take_slasher_snapshot,
    calculate_slasher_snapshot_difference,
    revert_slasher_snapshot,
)


def runtime_sync_hook_is_executing_extension(client):
    """
    Runtime sync hook to check whether a slash command should be registered and synced instantly when added or removed.
    
    Parameters
    ----------
    client : ``Client``
        The respective client of the ``Slasher``.
    """
    return not EXTENSION_LOADER.is_processing_extension()

RUNTIME_SYNC_HOOKS.append(runtime_sync_hook_is_executing_extension)



__all__ = ('Pagination', )

from ....backend.futures import CancelledError
from ....discord.interaction import ComponentButton, ComponentRow

from .menu import Menu
from .helpers import EMOJI_LEFT_2, EMOJI_LEFT, EMOJI_RIGHT, EMOJI_RIGHT_2, EMOJI_CANCEL, get_auto_check, \
    CUSTOM_ID_CANCEL, top_level_check, top_level_get_timeout

class Pagination(Menu):
    BUTTON_LEFT_2 = ComponentButton(emoji=EMOJI_LEFT_2)
    BUTTON_LEFT = ComponentButton(emoji=EMOJI_LEFT)
    BUTTON_RIGHT = ComponentButton(emoji=EMOJI_RIGHT)
    BUTTON_RIGHT_2 = ComponentButton(emoji=EMOJI_RIGHT_2)
    BUTTON_CANCEL = ComponentButton(emoji=EMOJI_CANCEL, custom_id=CUSTOM_ID_CANCEL)
    
    BUTTONS = ComponentRow(BUTTON_LEFT_2, BUTTON_LEFT, BUTTON_RIGHT, BUTTON_RIGHT_2, BUTTON_CANCEL,)
    
    __slots__ = ('page_index', 'pages', 'timeout', 'user_check')
    
    def __init__(self, client, event, pages, *, check=..., timeout=300.0):
        if check is ...:
            check = get_auto_check(event)
        
        self.pages = pages
        self.page_index = 0
        self.timeout = timeout
        self.user_check = check
    
    
    check = top_level_check
    get_timeout = top_level_get_timeout

    
    async def initial_invoke(self):
        self.components = self.BUTTONS
        self.allowed_mentions = None
        self.BUTTON_LEFT_2.enabled = False
        self.BUTTON_LEFT.enabled = False
        
        pages = self.pages
        self.content = pages[0]
        if len(pages) == 1:
            self.BUTTON_RIGHT_2.enabled = False
            self.BUTTON_RIGHT.enabled = False
   
    
    async def invoke(self, event):
        interaction = event.interaction
        if interaction == self.BUTTON_CANCEL:
            self.cancel(CancelledError())
            return False
        
        pages_index_limit = len(self.pages)-1
        
        if interaction == self.BUTTON_LEFT:
            page_index = self.page_index-1
        elif interaction == self.BUTTON_RIGHT:
            page_index = self.page_index+1
        elif interaction == self.BUTTON_LEFT_2:
            page_index = 0
        elif interaction == self.BUTTON_RIGHT_2:
            page_index = pages_index_limit
        else:
             return False
        
        if page_index < 0:
            page_index = 0
        elif page_index > pages_index_limit:
            page_index = pages_index_limit
        
        if self.page_index == page_index:
            return False
        
        if page_index == 0:
            self.BUTTON_LEFT_2.enabled = False
            self.BUTTON_LEFT.enabled = False
        else:
            self.BUTTON_LEFT_2.enabled = True
            self.BUTTON_LEFT.enabled = True
        
        if page_index == pages_index_limit:
            self.BUTTON_RIGHT_2.enabled = False
            self.BUTTON_RIGHT.enabled = False
        else:
            self.BUTTON_RIGHT_2.enabled = True
            self.BUTTON_RIGHT.enabled = True
        
        self.page_index = page_index
        self.content = self.pages[page_index]
        return True



__all__ = ('ThreadProfile', 'thread_user_create', 'thread_user_delete', 'thread_user_pop', 'thread_user_update')

from datetime import datetime

from ...backend.export import include

from ..utils import timestamp_to_datetime

from .flags import ThreadProfileFlag

create_partial_role_from_id = include('create_partial_role_from_id')
Client = include('Client')

def thread_user_create(thread_channel, user, thread_user_data):
    """
    Resolves the given thread user data.
    
    Parameters
    ----------
    thread_channel : ``ChannelThread``
        The respective thread.
    user : ``ClientUserBase``
        The respective user to add or update in the thread.
    thread_user_data : `dict` of (`str`, `Any`) items
        Received thread user data.
    
    Returns
    -------
    created : `bool`
        Whether a new thread profile was created.
    """
    thread_users = thread_channel.thread_users
    if thread_users is None:
        thread_users = thread_channel.thread_users = {}
    thread_users[user.id] = user
    
    thread_profiles = user.thread_profiles
    if thread_profiles is None:
        thread_profiles = user.thread_profiles = {}
    
    try:
        thread_profile = thread_profiles[thread_channel.id]
    except KeyError:
        thread_profiles[thread_channel.id] = ThreadProfile(thread_user_data)
        created = True
    else:
        thread_profile._update_attributes(thread_user_data)
        created = False
    
    return created


def thread_user_update(thread_channel, user, thread_user_data):
    """
    Resolves the given thread user update.
    
    Parameters
    ----------
    thread_channel : ``ChannelThread``
        The respective thread.
    user : ``ClientUserBase``
        The respective user to add or update in the thread.
    thread_user_data : `dict` of (`str`, `Any`) items
        Received thread user data.
    
    Returns
    -------
    old_attributes : `None` or `dict` of (`str`, `Any`) items
    """
    thread_users = thread_channel.thread_users
    if thread_users is None:
        thread_users = thread_channel.thread_users = {}
    thread_users[user.id] = user
    
    thread_profiles = user.thread_profiles
    if thread_profiles is None:
        thread_profiles = user.thread_profiles = {}
    
    try:
        thread_profile = thread_profiles[thread_channel.id]
    except KeyError:
        thread_profiles[thread_channel.id] = ThreadProfile(thread_user_data)
        return None
    
    old_attributes = thread_profile._update_attributes(thread_user_data)
    if not old_attributes:
        old_attributes = None
    
    return old_attributes


def thread_user_delete(thread_channel, user_id):
    """
    Removes the user for the given id from the thread's users.
    
    Parameters
    ----------
    thread_channel : ``ChannelThread``
        The respective thread.
    user_id : `int`
        The respective user's identifier.
    """
    thread_users = thread_channel.thread_users
    if (thread_users is not None):
        try:
            user = thread_users.pop(user_id)
        except KeyError:
            pass
        else:
            if not thread_users:
                thread_channel.thread_users = None
            
            thread_profiles = user.thread_profiles
            if (thread_profiles is not None):
                try:
                    del thread_profiles[thread_channel.id]
                except KeyError:
                    pass
                else:
                    if not thread_profiles:
                        user.thread_profiles = None


def thread_user_pop(thread_channel, user_id, me):
    """
    Removes and returns the user for the given id from the thread's users.
    
    Parameters
    ----------
    thread_channel : ``ChannelThread``
        The respective thread.
    user_id : `int`
        The respective user's identifier.
    me : ``Client``
        The client who pops the user.
    
    Returns
    -------
    popped : `None` or `tuple` (``ClientUserBase``, ``ThreadProfile``) item
        The removed user and it's profile if any.
    """
    thread_users = thread_channel.thread_users
    if (thread_users is not None):
        try:
            user = thread_users.pop(user_id)
        except KeyError:
            pass
        else:
            if not thread_users:
                thread_channel.thread_users = None
            
            thread_profiles = user.thread_profiles
            if (thread_profiles is not None):
                if isinstance(user, Client) and (user is not me):
                    thread_profile = thread_profiles.get(thread_channel.id, None)
                else:
                    try:
                        thread_profile = thread_profiles.pop(thread_channel.id)
                    except KeyError:
                        thread_profile = None
                    else:
                        if not thread_profiles:
                            user.thread_profiles = None
                        
                if (thread_profile is not None):
                    return user, thread_profile


class ThreadProfile:
    """
    Represents an user's profile inside of a thread channel.
    
    Attributes
    ----------
    joined_at : `datetime`
        The date when the user joined the thread.
    flags : ``ThreadProfileFlag``
        user specific settings of the profile.
    """
    __slots__ = ('joined_at', 'flags',)
    
    @property
    def created_at(self):
        """
        Returns ``.joined_at`` if set.
        
        Returns
        -------
        created_at : `datetime`
        """
        return self.joined_at
    
    def __init__(self, data):
        """
        Creates a new ``ThreadProfile`` instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received thread profile data.
        """
        self.joined_at = timestamp_to_datetime(data['join_timestamp'])
        
        self._update_attributes(data)
    
    def __repr__(self):
        """Returns the thread profile's representation."""
        return f'<{self.__class__.__name__}>'
    
    def _update_attributes(self, data):
        """
        Updates the thread profile with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received thread profile data.
        """
        self.flags = ThreadProfileFlag(data['flags'])
    
    def _difference_update_attributes(self, data):
        """
        Updates the thread profile and returns it's changed attributes in a `dict` within `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-------------------+-------------------------------+
        | Keys              | Values                        |
        +===================+===============================+
        | flags             | ``ThreadProfileFlag``         |
        +-------------------+-------------------------------+
        """
        old_attributes = {}
        
        flags = data.get('flags', 0)
        if self.flags != flags:
            old_attributes['flags'] = self.flags
            self.flags = ThreadProfileFlag(flags)
        
        return old_attributes



__all__ = ()

from ...backend.futures import Task

from ..exceptions import DiscordException
from ..core import KOKORO

from .core import PARSERS

SYNC_REQUESTS = {}

async def sync_task(queue_id, coro, queue):
    """
    Syncer task ensured if a guild related dispatch event fails, when any expected entity mentioned by it was not
    found.
    
    This function is a coroutine.
    
    Parameters
    ----------
    queue_id : `int`
        The respective guild's id to identify queued up unhandled dispatch event when de-sync happened.
    coro : `coroutine`
        ``Client.guild_sync`` coroutine.
    queue : `list` of `tuple` (``Client``, `Any`, (`str` or `tuple` (`str`, `function`, `Any`)))
        A queue of events to call with the specified parameters.
        
        First element of the queue is always the respective client of the received dispatch event. The second is the
        payload of the dispatch event, meanwhile the third can be the name of it (parser name case), or a `tuple` of
        3 elements (checker case), where the first is the name of the parser, the second is a checker and the third is
        an additional value to check with.
        
        At the parser name case, the parser will be called at every case, meanwhile at the checker case, the checker
        will be called with the synced guild and with the passed value, and then the parser will be called only, if the
        the checker returned `True`.
    """
    try:
        guild = await coro
    except (DiscordException, ConnectionError):
        return
    else:
        # Fix infinite loops by do not dispatching again on error
        for index in range(len(queue)):
            client, data, parser_and_checker = queue[index]
            if type(parser_and_checker) is str:
                PARSERS[parser_and_checker](client, data)
                continue
            
            parser_name, checker, value = parser_and_checker
            if checker(guild, value):
                PARSERS[parser_name](client, data)
    finally:
        del SYNC_REQUESTS[queue_id]


def check_channel(guild, channel_id):
    """
    Checks whether the given guild has a channel with the specified id.
    
    This function is a checker used at guild syncing.
    
    Parameters
    ----------
    guild : ``Guild``
    channel_id : `int`
    """
    return (channel_id in guild.channels)


def guild_sync(client, data, parser_and_checker):
    """
    Syncer function what is called when any expected entity mentioned by a dispatch event's parser was not found.
    
    Looks up whether the given guild has already a syncer task, if it has not, then creates a new ``sync_task`` for it.
    If `parser_and_checker` is given as not `None`, then the respective failed parser will be called when the syncer
    finished.
    
    Parameters
    ----------
    client : ``Client``
        The respective client of the dispatch event.
    data : `Any`
        The payload of the dispatch event.
    parser_and_checker : `None` or `str` or `tuple` (`str`, `function`, `Any`)
        - Is passed as `None` if only the syncer task should run.
        - Is passed as `str`, if the respective parser should be called when syncing is done.
        - Is passed as `tuple` of 3 elements : `str`, `function`, `Any`; if the respective parser's calling is bound to
            a condition. The passed `function` should contain the condition and accept the respective guild and the
            third value (the type `Any` one) as parameters and return the condition's result.
    """
    try:
        guild_id = int(data['guild_id'])
    except KeyError:
        return
    
    try:
        queue = SYNC_REQUESTS[guild_id]
    except KeyError:
        queue = []
        Task(sync_task(guild_id, client.guild_sync(guild_id), queue), KOKORO)
        SYNC_REQUESTS[guild_id] = queue
    
    if parser_and_checker is None:
        return
    queue.append((client, data, parser_and_checker),)



__all__ = ('Connection', )

from ..integration import Integration
from ..bases import DiscordEntity


class Connection(DiscordEntity):
    """
    A connection object that a user is attached to.
    
    Attributes
    ----------
    id : `int`
        The unique identifier value of the connection.
    friend_sync : `bool`
        Whether the user has friend sync enabled for the connection.
    integrations : `None` or (`list` of ``Integration``)
        A list (if any) of guild integrations which are attached to the connection.
    name : `str`
        The username of the connected account.
    revoked : `bool`
        Whether the connection is revoked.
    show_activity : `bool`
        Whether activity related to this connection will be shown in presence updates.
    type : `str`
        The service of the connection. (Like `'twitch'` or `'youtube'`.)
    verified : `bool`
        Whether the connection is verified.
    visibility : `int`
        For who is the connection visible for.
        
        Possible visibility values
        +-------+-------------------------------+
        | value | description                   |
        +=======+===============================+
        | 0     | Visible only for the user.    |
        +-------+-------------------------------+
        | 1     | Visible to everyone.          |
        +-------+-------------------------------+
    """
    __slots__ = ('friend_sync', 'integrations', 'name', 'revoked', 'show_activity', 'type', 'verified', 'visibility',)
    
    def __init__(self, data):
        """
        Creates a connection object from received connection data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`)
            Received connection data.
        """
        self.name = data['name']
        self.type = data['type']
        self.id = int(data['id'])
        self.revoked = data.get('revoked', False)
        self.verified = data.get('verified', False)
        self.show_activity = data.get('show_activity', False)
        self.friend_sync = data.get('friend_sync', False)
        self.visibility = data.get('visibility', 0)
        
        try:
            integration_datas = data['integrations']
        except KeyError:
            integrations = None
        else:
            if integration_datas:
                integrations = [Integration(integration_data) for integration_data in integration_datas]
            else:
                integrations = None
        self.integrations = integrations



"""
Hata extensions for supporting interactions.
"""
from .application_command import *
from .components import *
from .component_command import *
from .converters import *
from .event_handlers import *
from .exceptions import *
from .expression_parser import *
from .responding import *
from .slasher import *
from .utils import *
from .waiters import *
from .wrappers import *

__all__ = (
    'P',
    'configure_parameter',
    'set_permission',
    'setup_ext_slash',
    *application_command.__all__,
    *components.__all__,
    *component_command.__all__,
    *converters.__all__,
    *event_handlers.__all__,
    *exceptions.__all__,
    *expression_parser.__all__,
    *responding.__all__,
    *slasher.__all__,
    *utils.__all__,
    *waiters.__all__,
    *wrappers.__all__,
)

from .. import register_library_extension, add_library_extension_hook, register_setup_function

from .event_handlers import _do_initial_sync, _application_command_create_watcher, \
    _application_command_delete_watcher, _application_command_permission_update_watcher
from .client_wrapper_extension import *

set_permission = SlasherApplicationCommandPermissionOverwriteWrapper
configure_parameter = SlasherApplicationCommandParameterConfigurerWrapper
P = SlashParameter

def setup_ext_slash(client, **kwargs):
    """
    Setups the slash extension on client.
    
    Note, that this function can be called on a client only once.
    
    Parameters
    ----------
    client : ``Client``
        The client to setup the extension on.
    immediate_sync : `bool`, Optional
        Whether application command sync task should be started immediately as it is detected that no more features
        are added to the slasher. Defaults to `False`.
    **kwargs : Keyword parameters
        Additional keyword parameter to be passed to the created ``Slasher``.
    
    Other Parameters
    ----------------
    delete_commands_on_unload: `bool`, Optional
        Whether commands should be deleted when unloaded.
    use_default_exception_handler : `bool`, Optional
        Whether the default slash exception handler should be added as an exception handler.
    
    Returns
    -------
    slasher : ``Slasher``
        Slash command processor.
    
    Raises
    ------
    RuntimeError
        If the client has an attribute set what the slasher would use.
    TypeError
        - If `client` was not given as ``Client`` instance.
        - If `delete_commands_on_unload` was not given as `bool` instance.
        - If `use_default_exception_handler` was not given as `bool` instance.
    """
    for attr_name in ('slasher', 'interactions'):
        if hasattr(client, attr_name):
            raise RuntimeError(f'The client already has an attribute named as `{attr_name}`.')
    
    slasher = Slasher(client, **kwargs)
    
    client.events(slasher)
    client.slasher = slasher
    client.interactions = slasher.shortcut
    client.events(_do_initial_sync, name='launch')
    client.events(_application_command_create_watcher, name='application_command_create')
    client.events(_application_command_delete_watcher, name='application_command_delete')
    client.events(_application_command_permission_update_watcher, name='application_command_permission_update')
    
    return slasher


def snapshot_hook():
    from . import snapshot


register_library_extension('HuyaneMatsu.slash')
add_library_extension_hook(snapshot_hook, ['HuyaneMatsu.extension_loader'])

register_setup_function(
    'HuyaneMatsu.slash',
    setup_ext_slash,
    None,
    (
        'delete_commands_on_unload',
        'use_default_exception_handler',
        'random_error_message_getter',
    ),
)



__all__ = ()


def raw_name_to_display(raw_name):
    """
    Converts the given raw command or it's parameter's name to it's display name.
    
    Parameters
    ----------
    raw_name : `str`
        The name to convert.
    
    Returns
    -------
    display_name : `str`
        The converted name.
    """
    return '-'.join([w for w in raw_name.strip('_ ').lower().replace(' ', '-').replace('_', '-').split('-') if w])


def normalize_description(description):
    """
    Normalizes a docstrings.
    
    Parameters
    ----------
    description : `str` or `Any`
        The docstring to clear.
    
    Returns
    -------
    cleared : `str` or `Any`
        The cleared docstring. If `docstring` was given as `None` or is detected as empty, will return `None`.
    """
    if (description is None) or (not isinstance(description, str)):
        return description
    
    lines = description.splitlines()
    for index in reversed(range(len(lines))):
        line = lines[index]
        line = line.strip()
        if line:
            lines[index] = line
        else:
            del lines[index]
    
    if not lines:
        return None
    
    return ' '.join(lines)



__all__ = ('MessageIterator', )

from ..bases import maybe_snowflake
from ..core import CHANNELS
from ..exceptions import DiscordException, ERROR_CODES
from ..permission.permission import PERMISSION_MASK_READ_MESSAGE_HISTORY

from .channel_text_base import ChannelTextBase

# sounds funny, but this is a class
# the chunk_size is 99, because it means 1 request for _load_messages_till

class MessageIterator:
    """
    An asynchronous message iterator over the given text channel.
    
    Attributes
    ----------
    _index : `int`
        The index of the message, what will be yielded.
    _can_read_history : `bool`
        Tells the message iterator, whether it's client can read the history if it's channel.
    channel : ``ChannelTextBase`` instance
        The channel, what's messages the message iterator will iterates over.
    chunk_size : `int`
        The amount of messages, what the message iterator will extend it's channel's message history, each time, the
        loaded messages are exhausted.
    client : ``Client``
        The client, who will do the api requests for requesting more messages.
    """
    __slots__ = ('_can_read_history', '_index', 'channel', 'chunk_size', 'client',)
    
    async def __new__(cls, client, channel, chunk_size=99):
        """
        Creates a message iterator.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client, who will do the api requests for requesting more messages.
        channel : ``ChannelTextBase`` or `int` instance
            The channel, what's messages the message iterator will iterates over.
        chunk_size : `int`, Optional
            The amount of messages, what the message iterator will extend it's channel's message history, each time,
            the loaded messages are exhausted. Limited to `97` as a maximal value.
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``ChannelTextBase`` nor `int` instance.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `chunk_size` was not given as `int` instance.
            - If `chunk_size` is out of range [1:].
        """
        if __debug__:
            if not isinstance(chunk_size, int):
                raise AssertionError(f'`chunk_size` can be given as `int` instance, got '
                    f'{chunk_size.__class__.__name__}.')
            
            if chunk_size < 1:
                raise AssertionError(f'`chunk_size` is out from the expected [0:] range, got {chunk_size!r}.')
        
        if chunk_size > 99:
            chunk_size = 99
        
        if isinstance(channel, ChannelTextBase):
            pass
        else:
            channel_id = maybe_snowflake(channel)
            if channel_id is None:
                raise TypeError(f'`channel` can be given as `{ChannelTextBase.__name__}` instance, got'
                    f'{channel.__class__.__name__}.')
            
            channel = CHANNELS.get(channel_id, None)
            
            if channel is None:
                try:
                    messages = await client.message_get_chunk_from_zero(channel_id, 100)
                except BaseException as err:
                    if isinstance(err, DiscordException) and err.code in (
                            ERROR_CODES.unknown_message, # message deleted
                            ERROR_CODES.unknown_channel, # message's channel deleted
                            ERROR_CODES.missing_access, # client removed
                            ERROR_CODES.missing_permissions, # permissions changed meanwhile
                            ERROR_CODES.cannot_message_user, # user has dm-s disallowed
                                ):
                        pass
                    
                    else:
                        raise
                else:
                    if messages:
                        channel = messages[0].channel
                    else:
                        channel = None
        
        self = object.__new__(cls)
        self.client = client
        self.channel = channel
        self.chunk_size = chunk_size
        self._index = 0
        self._can_read_history = channel.cached_permissions_for(client)&PERMISSION_MASK_READ_MESSAGE_HISTORY
        return self
    
    def __aiter__(self):
        """Returns self and resets the `.index`."""
        self._index = 0
        return self
    
    async def __anext__(self):
        """
        Yields the next message of the iterator's channel.
        
        This method is a coroutine.
        """
        channel = self.channel
        if (channel is None):
            raise StopAsyncIteration
        
        index = self._index
        messages = channel.messages
        if (messages is not None) and (len(messages) > index):
            self._index = index+1
            return channel.messages[index]
        
        if channel.message_history_reached_end or (not self._can_read_history):
            raise StopAsyncIteration
        
        try:
            await self.client._load_messages_till(channel, index+self.chunk_size)
        except BaseException as err:
            if isinstance(err, ConnectionError):
                pass
            
            elif isinstance(err, DiscordException) and err.code in (
                ERROR_CODES.unknown_message, # message deleted
                ERROR_CODES.unknown_channel, # message's channel deleted
                ERROR_CODES.missing_access, # client removed
                ERROR_CODES.missing_permissions, # permissions changed meanwhile
                ERROR_CODES.cannot_message_user, # user has dm-s disallowed
            ):
                pass
            else:
                raise
        
        messages = channel.messages
        if (messages is not None) and (len(channel.messages) > index):
            self._index = index+1
            return channel.messages[index]
        
        raise StopAsyncIteration
    
    
    def __repr__(self):
        """Returns the representation of the message iterator."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' of client ', self.client.full_name,
        ]
        
        channel = self.channel
        if (channel is not None):
            repr_parts.append(', at channel ')
            repr_parts.append(repr(channel.name))
            repr_parts.append(' (')
            repr_parts.append(repr(channel.id))
            repr_parts.append(')')
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ('ScheduledEvent', )

from ..bases import DiscordEntity, IconSlot
from ..core import SCHEDULED_EVENTS, CHANNELS, GUILDS
from ..channel import CHANNEL_TYPES, create_partial_channel_from_id
from ..utils import timestamp_to_datetime

from .preinstanced import ScheduledEventStatus, ScheduledEventEntityType, PrivacyLevel

class ScheduledEvent(DiscordEntity):
    """
    Attributes
    ----------
    id : `int`
        The scheduled event's identifier number.
    channel_id : `int`
        The stage channel id of the event.
        
        Defaults to `0` if not applicable.
    description : `None` or `str`
        Description of the event.
    entity_id : `int`
        The event's entity's type.
        
        Defaults to `0`.
    entity_type : ``ScheduledEventEntityType``
        To which type of entity the event is bound to.
    guild_id : `int`
        The respective event's identifier.
    entity_metadata : `None` or ``ScheduledEventMetadata`` instance
        Metadata about the target entity.
    image_type : ``IconType``
        The event's image's type.
    image_hash : `int`
        The event's image hash.
    name : `str`
        The event's name.
    privacy_level : ``PrivacyLevel``
        The privacy level of the event.
    end : `None` or `datetime`
        The scheduled end time of the event.
    start : `None` or `datetime`
        The scheduled start time of the event.
    send_start_notification : `bool`
        Whether start notification should be sent when the event is started.
    sku_ids : `tuple` of `int`
        Stock keeping unit identifiers used at the event.
    status : ``ScheduledEventStatus``
        The status of the event.
    user_count : `int`
        Users subscribed to the event.
    """
    __slots__ = ('channel_id', 'description', 'entity_id', 'entity_metadata', 'entity_type', 'guild_id', 'name',
        'privacy_level', 'end', 'start', 'send_start_notification', 'sku_ids', 'status', 'user_count' )
    
    image = IconSlot('image', 'image', None, None, add_updater=False)
    
    def __new__(cls, data):
        """
        Creates a new ``ScheduledEvent`` instance from the received data.
        
        Ift he instance already exists, returns that instead.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Guild event data.
        """
        event_id = int(data['id'])
        
        try:
            self = SCHEDULED_EVENTS[event_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = event_id
            SCHEDULED_EVENTS[event_id] = self
            
            self._set_attributes(data)
        else:
            self._update_counts_only(data)
        
        return self
    
    
    def _set_attributes(self, data):
        """
        Sets the event's attributes form the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Scheduled event data.
        """
        self.user_count = data.get('user_count', 0)
        
        entity_type = ScheduledEventEntityType.get(data['entity_type'])
        self.entity_type = entity_type
        self.privacy_level = PrivacyLevel.get(data['privacy_level'])
        
        channel_id = data.get('channel_id', None)
        if channel_id is None:
            channel_id = 0
        else:
            channel_id = int(channel_id)
        self.channel_id = channel_id
        
        self.guild_id = int(data['guild_id'])
        
        entity_metadata_data = data.get('entity_metadata', None)
        if entity_metadata_data is None:
            entity_metadata = None
        else:
            metadata_type = entity_type.metadata_type
            if (metadata_type is None):
                entity_metadata = None
            else:
                entity_metadata = metadata_type.from_data(entity_metadata_data)
        
        self.entity_metadata = entity_metadata
        
        entity_id = data.get('entity_id', None)
        if entity_id is None:
            entity_id = 0
        else:
            entity_id = int(entity_id)
        self.entity_id = entity_id
        
        self._update_attributes(data)
    
    def _update_attributes(self, data):
        """
        Updates the attributes of the scheduled event.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Scheduled event data.
        """
        description = data.get('description', None)
        if (description is not None) and (not description):
            description = None
        
        self.description = description
        
        self.status = ScheduledEventStatus.get(data['status'])
        
        self.name = data['name']
        
        self._set_image(data)
        
        end = data.get('scheduled_end_time', None)
        if (end is not None):
            end = timestamp_to_datetime(end)
        self.end = end
        
        start = data.get('scheduled_start_time', None)
        if (start is not None):
            start = timestamp_to_datetime(start)
        self.start = start
        
        sku_ids = data.get('sku_ids')
        if (sku_ids is None) or (not sku_ids):
            sku_ids = None
        else:
            sku_ids = tuple(sorted(int(sku_id) for sku_id in sku_ids))
        self.sku_ids = sku_ids
        
        self.send_start_notification = data.get('send_start_notification', False)
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the attributes of the scheduled event and returns the changed ones within an `attribute-name` -
        `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Scheduled event data.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            The updated attributes.
            
            The returned dictionary might contain the following items:
            
            +---------------------------+-------------------------------+
            | Key                       | Value                         |
            +===========================+===============================+
            | description               | `None` or `str`               |
            +---------------------------+-------------------------------+
            | image                     | ``Icon``                      |
            +---------------------------+-------------------------------+
            | name                      | `str`                         |
            +---------------------------+-------------------------------+
            | send_start_notification   | `bool`                        |
            +---------------------------+-------------------------------+
            | end                       | `None` or `datetime`          |
            +---------------------------+-------------------------------+
            | start                     | `None` or `datetime`          |
            +---------------------------+-------------------------------+
            | sku_ids                   | `None` or `tuple` of `int`    |
            +---------------------------+-------------------------------+
            | status                    | ``ScheduledEventStatus``      |
            +---------------------------+-------------------------------+
        """
        old_attributes = {}
        
        description = data.get('description', None)
        if (description is not None) and (not description):
            description = None
        if self.description != description:
            old_attributes['description'] = self.description
            self.description = description
        
        status = ScheduledEventStatus.get(data['status'])
        if self.status is not status:
            old_attributes['status'] = self.status
            self.status = status
        
        name = data['name']
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        self._update_image(data, old_attributes)
        
        end = data.get('scheduled_end_time', None)
        if (end is not None):
            end = timestamp_to_datetime(end)
        if self.end != end:
            old_attributes['end'] = self.end
            self.end = end
        
        start = data.get('scheduled_start_time', None)
        if (start is not None):
            start = timestamp_to_datetime(start)
        if self.start != start:
            old_attributes['start'] = self.start
            self.start = start
        
        sku_ids = data.get('sku_ids')
        if (sku_ids is None) or (not sku_ids):
            sku_ids = None
        else:
            sku_ids = tuple(sorted(int(sku_id) for sku_id in sku_ids))
        if self.sku_ids != sku_ids:
            old_attributes['sku_ids'] = self.sku_ids
            self.sku_ids = sku_ids
        
        send_start_notification = data.get('send_start_notification', False)
        if self.send_start_notification != send_start_notification:
            old_attributes['send_start_notification'] = self.send_start_notification
            self.send_start_notification = send_start_notification
        
        return old_attributes
    
    def update_counts_only(self, data):
        """
        Updates the scheduled event's count attributes only.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Scheduled event data.
        """
        try:
            user_count = data['user_count']
        except KeyError:
            pass
        else:
            self.user_count = user_count
        
    
    def __repr__(self):
        """Returns the guild event's representation."""
        repr_parts = [
            '<', self.__class__.__name__,
            ' id=', repr(self.id),
            ', name=', repr(self.name),
            ', guild_id', repr(self.guild_id),
        ]
        
        entity_type = self.entity_type
        if entity_type is not ScheduledEventEntityType.none:
            repr_parts.append(', entity_type=')
            repr_parts.append(entity_type.name)
            repr_parts.append(' (')
            repr_parts.append(repr(entity_type.value))
            repr_parts.append(')')
        
        
        status = self.status
        if status is not ScheduledEventStatus.none:
            repr_parts.append(', status=')
            repr_parts.append(status.name)
            repr_parts.append(' (')
            repr_parts.append(repr(status.value))
            repr_parts.append(')')
        
        repr_parts.append('>')
        return ''.join(repr_parts)

    
    @property
    def entity(self):
        """
        Returns the stage channel's entity if applicable.
        
        Returns
        -------
        entity : `None` or ``ChannelStage``
        """
        entity_id = self.entity_id
        if entity_id:
            entity_type = self.entity_type
            if entity_type is ScheduledEventEntityType.none:
                entity = None
            elif entity_type is ScheduledEventEntityType.stage:
                entity = create_partial_channel_from_id(entity_id, CHANNEL_TYPES.guild_stage, self.guild_id)
            elif entity_type is ScheduledEventEntityType.voice:
                entity = create_partial_channel_from_id(entity_id, CHANNEL_TYPES.guild_voice, self.guild_id)
            else:
                entity = None
        else:
            entity = None
        
        return entity
    
    
    @property
    def channel(self):
        """
        Returns the event's channel if has any.
        
        Returns
        -------
        channel : `None` or ``ChannelStage``
        """
        channel_id = self.channel_id
        if channel_id:
            return CHANNELS.get(channel_id, None)
    
    
    @property
    def guild(self):
        """
        Returns the event's guild.
        
        Returns
        -------
        guild : `None` or ``Guild``
        """
        return GUILDS.get(self.guild_id, None)





__all__ = ('create_partial_user_from_id', )

from ...backend.export import export

from ..core import USERS

from .user import User

@export
def create_partial_user_from_id(user_id):
    """
    Creates a partial user from the given `user_id`. If the user already exists returns that instead.
    
    Parameters
    ----------
    user_id : `int`
        The unique identifier number of the user.
    
    Returns
    -------
    user : ``ClientUserBase``
    """
    try:
        return USERS[user_id]
    except KeyError:
        pass
    
    return User._create_empty(user_id)



from .client_gateway import *
from .heartbeat import *
from .rate_limit import *
from .voice_client_gateway import *

__all__ = (
    *client_gateway.__all__,
    *heartbeat.__all__,
    *rate_limit.__all__,
    *voice_client_gateway.__all__,
)



__all__ = ()

from time import perf_counter
from audioop import mul as audio_mul

from ...backend.futures import Task, Event, sleep, CancelledError

from ..core import KOKORO
from .opus import FRAME_LENGTH, SAMPLES_PER_FRAME

PLAYER_DELAY = FRAME_LENGTH/1000.0

del FRAME_LENGTH

class AudioPlayer:
    """
    Sends voice data through the voice client's socket.
    
    Attributes
    ----------
    done : `bool`
        Whether the audio player finished playing it's source.
    resumed_waiter : `threading.Event`
        Indicates whether the the audio player is not paused.
    source : ``AudioSource`` instance
        The audio source what the player reads each 20 ms.
    should_update : `bool`
        Whether the voice client should update itself.
    task : `None` or ``Task``
        Audio reader task. Set as `None` if the reader is stopped.
    voice_client : ``VoiceClient``
        The voice client of audio player.
    """
    __slots__ = ('done', 'resumed_waiter', 'should_update', 'source', 'task', 'voice_client')
    
    def __init__(self, voice_client, source):
        """
        Creates an starts the audio player.
        
        Parameters
        ----------
        voice_client : ``VoiceClient``
            The voice client of audio player.
        source : ``AudioSource`` instance
            The audio source what the player reads each 20 ms.
        """
        self.source = source
        self.voice_client = voice_client
        
        resumed_waiter = Event(KOKORO)
        resumed_waiter.set()    # we are not paused
        self.resumed_waiter = resumed_waiter
        self.should_update = True
        self.done = False
        
        self.task = Task(self.run(), KOKORO)
    
    async def run(self):
        """
        The main runner of ``AudioPlayer``. Is instantly started inside of a ``Task`` as the player is created.
        
        This method is a coroutine.
        """
        voice_client = self.voice_client
        start = perf_counter()
        loops = 0
        
        source = None
        
        try:
            while True:
                if self.should_update:
                    source = await self.update(source)
                    if source is None:
                        break
                    
                    start = perf_counter()
                    loops = 0
                    continue
                
                # are we disconnected from voice?
                if not voice_client.connected.is_set():
                    await voice_client.connected
                    start = perf_counter()
                    loops = 0
                    continue
                
                loops += 1
                
                data = await source.read()
                
                if data is None:
                    self.source = None
                    await source.cleanup()
                    self.pause()
                    
                    async with voice_client.lock:
                        await voice_client.call_after(voice_client, source)
                    
                    source = None
                    self.should_update = True # safety first
                    continue
                
                sequence = voice_client._sequence
                if sequence == 65535:
                    sequence = 0
                else:
                    sequence += 1
                voice_client._sequence = sequence
                
                if source.NEEDS_ENCODE:
                    pref_volume = voice_client._pref_volume
                    if (pref_volume != 1.0):
                        data = audio_mul(data, 2, pref_volume)
                    
                    data = voice_client._encoder.encode(data)
                
                header = b''.join([
                    b'\x80x',
                    voice_client._sequence.to_bytes(2, 'big'),
                    voice_client._timestamp.to_bytes(4, 'big'),
                    voice_client._audio_source.to_bytes(4, 'big'),
                ])
                
                nonce = header+b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
                packet = bytearray(header)+voice_client._secret_box.encrypt(bytes(data), nonce).ciphertext
                
                voice_client.send_packet(packet)
                
                timestamp = voice_client._timestamp+SAMPLES_PER_FRAME
                if timestamp > 4294967295:
                    timestamp = 0
                voice_client._timestamp = timestamp
                
                delay = PLAYER_DELAY+((start+PLAYER_DELAY*loops)-perf_counter())
                if delay < 0.0:
                    continue
                await sleep(delay, KOKORO)
        
        except BaseException as err:
            if voice_client.player is self:
                voice_client.player = None
            
            self.done = True
            
            self.source = None
            if (source is not None):
                await source.cleanup()
                source = None
            
            if isinstance(err, GeneratorExit):
                raise
            
            if isinstance(err, CancelledError):
                return
            
            await KOKORO.render_exc_async(err, before=[
                'Exception occurred at \n',
                repr(self),
                '\n',
            ])
            
        else:
            if voice_client.player is self:
                voice_client.player = None
        
        finally:
            self.task = None
            
            # Force resume if applicable.
            if voice_client.player is None:
                queue = voice_client.queue
                if queue:
                    voice_client.player = type(self)(voice_client, queue.pop(0))
    
    async def update(self, actual_source):
        """
        Updates the player if ``.should_update`` is set as `True`.
        
        Waits for it's resumed waiter to be set if paused. If the voice player's source is updated, then initializes it
        as well and closes the old one too.
        
        This method is a coroutine.
        
        Parameters
        ----------
        actual_source : `None` or ``AudioSource`` instance
            The actual audio source of the player.
        
        Returns
        -------
        new_source : `None` or `AudioSource`` instance
            New source of the player to play.
            
            Can be same as the `actual_source`.
        """
        resumed_waiter = self.resumed_waiter
        if not resumed_waiter.is_set():
            await resumed_waiter
        
        self.should_update = False
        new_source = self.source
        if (new_source is None):
            if (self.voice_client.player is not self):
                self.done = True
                return None
            
            if (actual_source is not None):
                await actual_source.cleanup()
            
            return None
        
        if (new_source is actual_source):
            return actual_source
        
        if (actual_source is not None):
            await actual_source.cleanup()
        
        await new_source.postprocess()
        return new_source
    
    def pause(self):
        """
        Pauses the player.
        """
        self.resumed_waiter.clear()
        self.should_update = True
    
    def resume(self):
        """
        Resumes the player if paused.
        """
        resumed_waiter = self.resumed_waiter
        if not resumed_waiter.is_set():
            resumed_waiter.set()
            self.should_update = True
    
    def stop(self):
        """
        Stops the player if running.
        """
        if self.done:
            return
        
        self.done = True
        task = self.task
        if (task is not None):
            task.cancel()
        
        self.should_update = True
        self.resumed_waiter.set()
    
    def set_source(self, source):
        """
        Sets they player source.
        
        Parameters
        ----------
        source : `None` or ``AudioSource`` instance
            The new source of the player.
        """
        self.source = source
        self.should_update = True
        
        resumed_waiter = self.resumed_waiter
        if not resumed_waiter.is_set():
            resumed_waiter.set()



from .application_command import *
from .components import *
from .event_types import *
from .preinstanced import *
from .interaction_response_types import *
from . import interaction_response_types as INTERACTION_RESPONSE_TYPES

__all__ = (
    'INTERACTION_RESPONSE_TYPES',
    *application_command.__all__,
    *components.__all__,
    *event_types.__all__,
    *interaction_response_types.__all__,
    *preinstanced.__all__,
)
